from __future__ import annotations

from abc import abstractmethod
from datetime import datetime
from typing import TYPE_CHECKING, Sequence, cast, overload

from ._generated_api_client.api import (
    create_finetuned_embedding_model,
    delete_finetuned_embedding_model,
    embed_with_finetuned_model_gpu,
    embed_with_pretrained_model_gpu,
    get_finetuned_embedding_model,
    get_pretrained_embedding_model,
    list_finetuned_embedding_models,
    list_pretrained_embedding_models,
)
from ._generated_api_client.models import (
    EmbeddingFinetuningMethod,
    EmbedRequest,
    FinetunedEmbeddingModelMetadata,
    FinetuneEmbeddingModelRequest,
    FinetuneEmbeddingModelRequestTrainingArgs,
    PretrainedEmbeddingModelMetadata,
    PretrainedEmbeddingModelName,
)
from ._utils.common import CreateMode, DropMode
from ._utils.task import TaskStatus, wait_for_task
from .datasource import Datasource

if TYPE_CHECKING:
    from .memoryset import LabeledMemoryset


class _EmbeddingModel:
    name: str
    embedding_dim: int
    max_seq_length: int
    uses_context: bool

    def __init__(self, *, name: str, embedding_dim: int, max_seq_length: int, uses_context: bool):
        self.name = name
        self.embedding_dim = embedding_dim
        self.max_seq_length = max_seq_length
        self.uses_context = uses_context

    @classmethod
    @abstractmethod
    def all(cls) -> Sequence[_EmbeddingModel]:
        pass

    @overload
    def embed(self, value: str, max_seq_length: int | None = None) -> list[float]:
        pass

    @overload
    def embed(self, value: list[str], max_seq_length: int | None = None) -> list[list[float]]:
        pass

    def embed(self, value: str | list[str], max_seq_length: int | None = None) -> list[float] | list[list[float]]:
        """
        Generate embeddings for a value or list of values

        Params:
            value: The value or list of values to embed
            max_seq_length: The maximum sequence length to truncate the input to

        Returns:
            A matrix of floats representing the embedding for each value if the input is a list of
                values, or a list of floats representing the embedding for the single value if the
                input is a single value
        """
        request = EmbedRequest(values=value if isinstance(value, list) else [value], max_seq_length=max_seq_length)
        if isinstance(self, PretrainedEmbeddingModel):
            embeddings = embed_with_pretrained_model_gpu(self._model_name, body=request)
        elif isinstance(self, FinetunedEmbeddingModel):
            embeddings = embed_with_finetuned_model_gpu(self.id, body=request)
        else:
            raise ValueError("Invalid embedding model")
        return embeddings if isinstance(value, list) else embeddings[0]


class _PretrainedEmbeddingModelMeta(type):
    def __getattr__(cls, name: str) -> PretrainedEmbeddingModel:
        if cls != FinetunedEmbeddingModel and name in PretrainedEmbeddingModelName.__members__:
            return PretrainedEmbeddingModel._get(name)
        else:
            raise AttributeError(f"'{cls.__name__}' object has no attribute '{name}'")


class PretrainedEmbeddingModel(_EmbeddingModel, metaclass=_PretrainedEmbeddingModelMeta):
    """
    A pretrained embedding model

    **Models:**

    OrcaCloud supports a select number of small to medium sized embedding models that perform well on the
        [Hugging Face MTEB Leaderboard](https://huggingface.co/spaces/mteb/leaderboard).
        These can be accessed as class attributes. We currently support:

    - **`CDE_SMALL`**: Context-aware CDE small model from Hugging Face ([jxm/cde-small-v1](https://huggingface.co/jxm/cde-small-v1))
    - **`CLIP_BASE`**: Multi-modal CLIP model from Hugging Face ([sentence-transformers/clip-ViT-L-14](https://huggingface.co/sentence-transformers/clip-ViT-L-14))
    - **`GTE_BASE`**: Alibaba's GTE model from Hugging Face ([Alibaba-NLP/gte-base-en-v1.5](https://huggingface.co/Alibaba-NLP/gte-base-en-v1.5))

    Examples:
        >>> PretrainedEmbeddingModel.CDE_SMALL
        PretrainedEmbeddingModel({name: CDE_SMALL, embedding_dim: 768, max_seq_length: 512})

    Attributes:
        name: Name of the pretrained embedding model
        embedding_dim: Dimension of the embeddings that are generated by the model
        max_seq_length: Maximum input length (in tokens not characters) that this model can process. Inputs that are longer will be truncated during the embedding process
        uses_context: Whether the pretrained embedding model uses context
    """

    _model_name: PretrainedEmbeddingModelName

    def __init__(self, metadata: PretrainedEmbeddingModelMetadata):
        # for internal use only, do not document
        self._model_name = metadata.name
        super().__init__(
            name=metadata.name.value,
            embedding_dim=metadata.embedding_dim,
            max_seq_length=metadata.max_seq_length,
            uses_context=metadata.uses_context,
        )

    def __eq__(self, other) -> bool:
        return isinstance(other, PretrainedEmbeddingModel) and self.name == other.name

    def __repr__(self) -> str:
        return f"PretrainedEmbeddingModel({{name: {self.name}, embedding_dim: {self.embedding_dim}, max_seq_length: {self.max_seq_length}}})"

    @classmethod
    def all(cls) -> list[PretrainedEmbeddingModel]:
        """
        List all pretrained embedding models in the OrcaCloud

        Returns:
            A list of all pretrained embedding models available in the OrcaCloud
        """
        return [cls(metadata) for metadata in list_pretrained_embedding_models()]

    _instances: dict[str, PretrainedEmbeddingModel] = {}

    @classmethod
    def _get(cls, name: PretrainedEmbeddingModelName | str) -> PretrainedEmbeddingModel:
        # for internal use only, do not document - we want people to use dot notation to get the model
        if str(name) not in cls._instances:
            cls._instances[str(name)] = cls(get_pretrained_embedding_model(cast(PretrainedEmbeddingModelName, name)))
        return cls._instances[str(name)]

    @classmethod
    def exists(cls, name: str) -> bool:
        """
        Check if a pretrained embedding model exists by name

        Params:
            name: The name of the pretrained embedding model

        Returns:
            True if the pretrained embedding model exists, False otherwise
        """
        return name in PretrainedEmbeddingModelName

    def finetune(
        self,
        name: str,
        train_datasource: Datasource | LabeledMemoryset,
        *,
        eval_datasource: Datasource | None = None,
        label_column: str = "label",
        value_column: str = "value",
        training_method: EmbeddingFinetuningMethod | str = EmbeddingFinetuningMethod.CLASSIFICATION,
        training_args: dict | None = None,
        if_exists: CreateMode = "error",
    ) -> FinetunedEmbeddingModel:
        """
        Finetune an embedding model

        Params:
            name: Name of the finetuned embedding model
            train_datasource: Data to train on
            eval_datasource: Optionally provide data to evaluate on
            label_column: Column name of the label
            value_column: Column name of the value
            training_method: Training method to use
            training_args: Optional override for Hugging Face [`TrainingArguments`](transformers.TrainingArguments).
                If not provided, reasonable training arguments will be used for the specified training method
            if_exists: What to do if a finetuned embedding model with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing finetuned embedding model.

        Returns:
            The finetuned embedding model

        Raises:
            ValueError: If the finetuned embedding model already exists and `if_exists` is `"error"` or if it is `"open"`
                but the base model param does not match the existing model

        Examples:
            >>> datasource = Datasource.open("my_datasource")
            >>> model = PretrainedEmbeddingModel.CLIP_BASE
            >>> model.finetune("my_finetuned_model", datasource)
        """
        exists = FinetunedEmbeddingModel.exists(name)

        if exists and if_exists == "error":
            raise ValueError(f"Finetuned embedding model '{name}' already exists")
        elif exists and if_exists == "open":
            existing = FinetunedEmbeddingModel.open(name)

            if existing.base_model_name != self._model_name:
                raise ValueError(f"Finetuned embedding model '{name}' already exists, but with different base model")

            return existing

        from .memoryset import LabeledMemoryset

        train_datasource_id = train_datasource.id if isinstance(train_datasource, Datasource) else None
        train_memoryset_id = train_datasource.id if isinstance(train_datasource, LabeledMemoryset) else None
        assert train_datasource_id is not None or train_memoryset_id is not None
        res = create_finetuned_embedding_model(
            body=FinetuneEmbeddingModelRequest(
                name=name,
                base_model=self._model_name,
                train_memoryset_id=train_memoryset_id,
                train_datasource_id=train_datasource_id,
                eval_datasource_id=eval_datasource.id if eval_datasource is not None else None,
                label_column=label_column,
                value_column=value_column,
                training_method=EmbeddingFinetuningMethod(training_method),
                training_args=(FinetuneEmbeddingModelRequestTrainingArgs.from_dict(training_args or {})),
            ),
        )
        wait_for_task(res.finetuning_task_id, description="Finetuning embedding model")
        return FinetunedEmbeddingModel.open(res.id)


class FinetunedEmbeddingModel(_EmbeddingModel):
    """
    A finetuned embedding model in the OrcaCloud

    Attributes:
        name: Name of the finetuned embedding model
        embedding_dim: Dimension of the embeddings that are generated by the model
        max_seq_length: Maximum input length (in tokens not characters) that this model can process. Inputs that are longer will be truncated during the embedding process
        uses_context: Whether the model uses the memoryset to contextualize embeddings (acts akin to inverse document frequency in TFIDF features)
        id: Unique identifier of the finetuned embedding model
        base_model: Base model the finetuned embedding model was trained on
        created_at: When the model was finetuned
    """

    id: str
    created_at: datetime
    updated_at: datetime
    _status: TaskStatus

    def __init__(self, metadata: FinetunedEmbeddingModelMetadata):
        # for internal use only, do not document
        self.id = metadata.id
        self.created_at = metadata.created_at
        self.updated_at = metadata.updated_at
        self.base_model_name = metadata.base_model
        self._status = metadata.finetuning_status
        super().__init__(
            name=metadata.name,
            embedding_dim=metadata.embedding_dim,
            max_seq_length=metadata.max_seq_length,
            uses_context=metadata.uses_context,
        )

    def __eq__(self, other) -> bool:
        return isinstance(other, FinetunedEmbeddingModel) and self.id == other.id

    def __repr__(self) -> str:
        return (
            "FinetunedEmbeddingModel({\n"
            f"    name: {self.name},\n"
            f"    embedding_dim: {self.embedding_dim},\n"
            f"    max_seq_length: {self.max_seq_length},\n"
            f"    base_model: PretrainedEmbeddingModel.{self.base_model_name.value}\n"
            "})"
        )

    @property
    def base_model(self) -> PretrainedEmbeddingModel:
        """Pretrained model the finetuned embedding model was based on"""
        return PretrainedEmbeddingModel._get(self.base_model_name)

    @classmethod
    def all(cls) -> list[FinetunedEmbeddingModel]:
        """
        List all finetuned embedding model handles in the OrcaCloud

        Returns:
            A list of all finetuned embedding model handles in the OrcaCloud
        """
        return [cls(metadata) for metadata in list_finetuned_embedding_models()]

    @classmethod
    def open(cls, name: str) -> FinetunedEmbeddingModel:
        """
        Get a handle to a finetuned embedding model in the OrcaCloud

        Params:
            name: The name or unique identifier of a finetuned embedding model

        Returns:
            A handle to the finetuned embedding model in the OrcaCloud

        Raises:
            LookupError: If the finetuned embedding model does not exist
        """
        return cls(get_finetuned_embedding_model(name))

    @classmethod
    def exists(cls, name_or_id: str) -> bool:
        """
        Check if a finetuned embedding model with the given name or id exists.

        Params:
            name_or_id: The name or id of the finetuned embedding model

        Returns:
            True if the finetuned embedding model exists, False otherwise
        """
        try:
            cls.open(name_or_id)
            return True
        except LookupError:
            return False

    @classmethod
    def drop(cls, name_or_id: str, *, if_not_exists: DropMode = "error"):
        """
        Delete the finetuned embedding model from the OrcaCloud

        Params:
            name_or_id: The name or id of the finetuned embedding model

        Raises:
            LookupError: If the finetuned embedding model does not exist and `if_not_exists` is `"error"`
        """
        try:
            delete_finetuned_embedding_model(name_or_id)
        except LookupError:
            if if_not_exists == "error":
                raise



"""
This file is generated by the openapi-python-client tool via the generate_api_client.py script

It is a customized template from the openapi-python-client tool's default template:
https://github.com/openapi-generators/openapi-python-client/blob/861ef5622f10fc96d240dc9becb0edf94e61446c/openapi_python_client/templates/endpoint_module.py.jinja

The main changes are:
- Update the API call responses to either return the successful response type or raise an error by:
    - Updating the _parse_response function to raise an error if the response status code is not in the 2xx range
- Inject a client into every method via a context manager
"""

from http import HTTPStatus
from typing import Any

import httpx

from ...client import _client_context
from ...errors import get_error_for_response
from ...types import Response


def _get_kwargs() -> dict[str, Any]:
    _kwargs: dict[str, Any] = {
        "method": "get",
        "url": "/",
    }

    return _kwargs


def _parse_response(*, response: httpx.Response) -> None:
    if response.status_code == 200:
        response_200 = response.json()
        return response_200
    if response.status_code == 500:
        raise get_error_for_response(response)
    if response.status_code == 503:
        raise get_error_for_response(response)
    else:
        raise get_error_for_response(response)


def _build_response(*, response: httpx.Response) -> Response[None]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(response=response),
    )


def sync_detailed() -> Response[None]:
    """Healthcheck

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[None]
    """

    kwargs = _get_kwargs()

    with _client_context() as client:
        response = client.get_httpx_client().request(
            **kwargs,
        )

        return _build_response(response=response)


def sync() -> None:
    """Healthcheck

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        None
    """

    return sync_detailed().parsed


async def asyncio_detailed() -> Response[None]:
    """Healthcheck

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[None]
    """

    kwargs = _get_kwargs()

    with _client_context() as client:
        response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(response=response)


async def asyncio() -> None:
    """Healthcheck

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        None
    """

    return (await asyncio_detailed()).parsed



"""
This file is generated by the openapi-python-client tool via the generate_api_client.py script

It is a customized template from the openapi-python-client tool's default template:
https://github.com/openapi-generators/openapi-python-client/blob/861ef5622f10fc96d240dc9becb0edf94e61446c/openapi_python_client/templates/model.py.jinja

The main change is:
- Fix typing issues
"""

# flake8: noqa: C901

from enum import Enum
from typing import Any, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.pretrained_embedding_model_name import PretrainedEmbeddingModelName
from ..types import UNSET, Unset

T = TypeVar("T", bound="EmbeddingEvaluationRequest")


@_attrs_define
class EmbeddingEvaluationRequest:
    """
    Attributes:
        value_column (str):
        label_column (str):
        source_id_column (Union[None, str]):
        neighbor_count (Union[Unset, int]):  Default: 5.
        label_names (Union[List[str], None, Unset]):
        embedding_models (Union[List[PretrainedEmbeddingModelName], None, Unset]):
    """

    value_column: str
    label_column: str
    source_id_column: Union[None, str]
    neighbor_count: Union[Unset, int] = 5
    label_names: Union[List[str], None, Unset] = UNSET
    embedding_models: Union[List[PretrainedEmbeddingModelName], None, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        value_column = self.value_column

        label_column = self.label_column

        source_id_column: Union[None, str]
        source_id_column = self.source_id_column

        neighbor_count = self.neighbor_count

        label_names: Union[List[str], None, Unset]
        if isinstance(self.label_names, Unset):
            label_names = UNSET
        elif isinstance(self.label_names, list):
            label_names = self.label_names

        else:
            label_names = self.label_names

        embedding_models: Union[List[str], None, Unset]
        if isinstance(self.embedding_models, Unset):
            embedding_models = UNSET
        elif isinstance(self.embedding_models, list):
            embedding_models = []
            for embedding_models_type_0_item_data in self.embedding_models:
                embedding_models_type_0_item = (
                    embedding_models_type_0_item_data.value
                    if isinstance(embedding_models_type_0_item_data, Enum)
                    else embedding_models_type_0_item_data
                )
                embedding_models.append(embedding_models_type_0_item)

        else:
            embedding_models = self.embedding_models

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "value_column": value_column,
                "label_column": label_column,
                "source_id_column": source_id_column,
            }
        )
        if neighbor_count is not UNSET:
            field_dict["neighbor_count"] = neighbor_count
        if label_names is not UNSET:
            field_dict["label_names"] = label_names
        if embedding_models is not UNSET:
            field_dict["embedding_models"] = embedding_models

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        value_column = d.pop("value_column")

        label_column = d.pop("label_column")

        def _parse_source_id_column(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        source_id_column = _parse_source_id_column(d.pop("source_id_column"))

        neighbor_count = d.pop("neighbor_count", UNSET)

        def _parse_label_names(data: object) -> Union[List[str], None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                label_names_type_0 = cast(List[str], data)

                return label_names_type_0
            except:  # noqa: E722
                pass
            return cast(Union[List[str], None, Unset], data)

        label_names = _parse_label_names(d.pop("label_names", UNSET))

        def _parse_embedding_models(data: object) -> Union[List[PretrainedEmbeddingModelName], None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                embedding_models_type_0 = []
                _embedding_models_type_0 = data
                for embedding_models_type_0_item_data in _embedding_models_type_0:
                    embedding_models_type_0_item = PretrainedEmbeddingModelName(embedding_models_type_0_item_data)

                    embedding_models_type_0.append(embedding_models_type_0_item)

                return embedding_models_type_0
            except:  # noqa: E722
                pass
            return cast(Union[List[PretrainedEmbeddingModelName], None, Unset], data)

        embedding_models = _parse_embedding_models(d.pop("embedding_models", UNSET))

        embedding_evaluation_request = cls(
            value_column=value_column,
            label_column=label_column,
            source_id_column=source_id_column,
            neighbor_count=neighbor_count,
            label_names=label_names,
            embedding_models=embedding_models,
        )

        embedding_evaluation_request.additional_properties = d
        return embedding_evaluation_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from uuid import uuid4

import pytest
from datasets.arrow_dataset import Dataset

from .classification_model import ClassificationModel
from .datasource import Datasource
from .embedding_model import PretrainedEmbeddingModel
from .memoryset import LabeledMemoryset


def test_create_model(model: ClassificationModel, memoryset: LabeledMemoryset):
    assert model is not None
    assert model.name == "test_model"
    assert model.memoryset == memoryset
    assert model.num_classes == 2
    assert model.memory_lookup_count == 3


def test_create_model_already_exists_error(memoryset, model: ClassificationModel):
    with pytest.raises(ValueError):
        ClassificationModel.create("test_model", memoryset)
    with pytest.raises(ValueError):
        ClassificationModel.create("test_model", memoryset, if_exists="error")


def test_create_model_already_exists_return(memoryset, model: ClassificationModel):
    with pytest.raises(ValueError):
        ClassificationModel.create("test_model", memoryset, if_exists="open", head_type="MMOE")

    with pytest.raises(ValueError):
        ClassificationModel.create("test_model", memoryset, if_exists="open", memory_lookup_count=37)

    with pytest.raises(ValueError):
        ClassificationModel.create("test_model", memoryset, if_exists="open", num_classes=19)

    with pytest.raises(ValueError):
        ClassificationModel.create("test_model", memoryset, if_exists="open", min_memory_weight=0.77)

    new_model = ClassificationModel.create("test_model", memoryset, if_exists="open")
    assert new_model is not None
    assert new_model.name == "test_model"
    assert new_model.memoryset == memoryset
    assert new_model.num_classes == 2
    assert new_model.memory_lookup_count == 3


def test_create_model_unauthenticated(unauthenticated, memoryset: LabeledMemoryset):
    with pytest.raises(ValueError, match="Invalid API key"):
        ClassificationModel.create("test_model", memoryset)


def test_get_model(model: ClassificationModel):
    fetched_model = ClassificationModel.open(model.name)
    assert fetched_model is not None
    assert fetched_model.id == model.id
    assert fetched_model.name == model.name
    assert fetched_model.num_classes == 2
    assert fetched_model.memory_lookup_count == 3
    assert fetched_model == model


def test_get_model_unauthenticated(unauthenticated):
    with pytest.raises(ValueError, match="Invalid API key"):
        ClassificationModel.open("test_model")


def test_get_model_invalid_input():
    with pytest.raises(ValueError, match="Invalid input"):
        ClassificationModel.open("not valid id")


def test_get_model_not_found():
    with pytest.raises(LookupError):
        ClassificationModel.open(str(uuid4()))


def test_get_model_unauthorized(unauthorized, model: ClassificationModel):
    with pytest.raises(LookupError):
        ClassificationModel.open(model.name)


def test_list_models(model: ClassificationModel):
    models = ClassificationModel.all()
    assert len(models) > 0
    assert any(model.name == model.name for model in models)


def test_list_models_unauthenticated(unauthenticated):
    with pytest.raises(ValueError, match="Invalid API key"):
        ClassificationModel.all()


def test_list_models_unauthorized(unauthorized, model: ClassificationModel):
    assert ClassificationModel.all() == []


def test_delete_model(memoryset: LabeledMemoryset):
    ClassificationModel.create("model_to_delete", LabeledMemoryset.open(memoryset.name))
    assert ClassificationModel.open("model_to_delete")
    ClassificationModel.drop("model_to_delete")
    with pytest.raises(LookupError):
        ClassificationModel.open("model_to_delete")


def test_delete_model_unauthenticated(unauthenticated, model: ClassificationModel):
    with pytest.raises(ValueError, match="Invalid API key"):
        ClassificationModel.drop(model.name)


def test_delete_model_not_found():
    with pytest.raises(LookupError):
        ClassificationModel.drop(str(uuid4()))
    # ignores error if specified
    ClassificationModel.drop(str(uuid4()), if_not_exists="ignore")


def test_delete_model_unauthorized(unauthorized, model: ClassificationModel):
    with pytest.raises(LookupError):
        ClassificationModel.drop(model.name)


def test_delete_memoryset_before_model_constraint_violation(hf_dataset):
    memoryset = LabeledMemoryset.from_hf_dataset("test_memoryset_delete_before_model", hf_dataset, value_column="text")
    ClassificationModel.create("test_model_delete_before_memoryset", memoryset)
    with pytest.raises(RuntimeError):
        LabeledMemoryset.drop(memoryset.id)


def test_evaluate(model):
    eval_datasource = Datasource.from_list(
        "eval_datasource",
        [
            {"text": "chicken noodle soup is the best", "label": 1},
            {"text": "cats are cute", "label": 0},
            {"text": "soup is great for the winter", "label": 0},
            {"text": "i love cats", "label": 1},
        ],
    )
    result = model.evaluate(eval_datasource, value_column="text")
    assert result is not None
    assert isinstance(result, dict)
    assert isinstance(result["accuracy"], float)
    assert isinstance(result["f1_score"], float)
    assert isinstance(result["loss"], float)
    assert len(result["precision_recall_curve"]["thresholds"]) == 4
    assert len(result["precision_recall_curve"]["precisions"]) == 4
    assert len(result["precision_recall_curve"]["recalls"]) == 4
    assert len(result["roc_curve"]["thresholds"]) == 4
    assert len(result["roc_curve"]["false_positive_rates"]) == 4
    assert len(result["roc_curve"]["true_positive_rates"]) == 4


def test_evaluate_with_telemetry(model):
    samples = [
        {"text": "chicken noodle soup is the best", "label": 1},
        {"text": "cats are cute", "label": 0},
    ]
    eval_datasource = Datasource.from_list("eval_datasource_2", samples)
    result = model.evaluate(eval_datasource, value_column="text", record_predictions=True, tags={"test"})
    assert result is not None
    predictions = model.predictions(tag="test")
    assert len(predictions) == 2
    assert all(p.tags == {"test"} for p in predictions)
    assert all(p.expected_label == s["label"] for p, s in zip(predictions, samples))


def test_predict(model: ClassificationModel, label_names: list[str]):
    predictions = model.predict(["Do you love soup?", "Are cats cute?"])
    assert len(predictions) == 2
    assert predictions[0].label == 0
    assert predictions[0].label_name == label_names[0]
    assert 0 <= predictions[0].confidence <= 1
    assert predictions[1].label == 1
    assert predictions[1].label_name == label_names[1]
    assert 0 <= predictions[1].confidence <= 1


def test_predict_unauthenticated(unauthenticated, model: ClassificationModel):
    with pytest.raises(ValueError, match="Invalid API key"):
        model.predict(["Do you love soup?", "Are cats cute?"])


def test_predict_unauthorized(unauthorized, model: ClassificationModel):
    with pytest.raises(LookupError):
        model.predict(["Do you love soup?", "Are cats cute?"])


def test_predict_constraint_violation(memoryset: LabeledMemoryset):
    model = ClassificationModel.create(
        "test_model_lookup_count_too_high", memoryset, num_classes=2, memory_lookup_count=memoryset.length + 2
    )
    with pytest.raises(RuntimeError):
        model.predict("test")


def test_record_prediction_feedback(model: ClassificationModel):
    predictions = model.predict(["Do you love soup?", "Are cats cute?"])
    expected_labels = [0, 1]
    model.record_feedback(
        {
            "prediction_id": p.prediction_id,
            "category": "correct",
            "value": p.label == expected_label,
        }
        for expected_label, p in zip(expected_labels, predictions)
    )


def test_record_prediction_feedback_missing_category(model: ClassificationModel):
    prediction = model.predict("Do you love soup?")
    with pytest.raises(ValueError):
        model.record_feedback({"prediction_id": prediction.prediction_id, "value": True})


def test_record_prediction_feedback_invalid_value(model: ClassificationModel):
    prediction = model.predict("Do you love soup?")
    with pytest.raises(ValueError, match=r"Invalid input.*"):
        model.record_feedback({"prediction_id": prediction.prediction_id, "category": "correct", "value": "invalid"})


def test_record_prediction_feedback_invalid_prediction_id(model: ClassificationModel):
    with pytest.raises(ValueError, match=r"Invalid input.*"):
        model.record_feedback({"prediction_id": "invalid", "category": "correct", "value": True})


def test_predict_with_memoryset_override(model: ClassificationModel, hf_dataset: Dataset):
    inverted_labeled_memoryset = LabeledMemoryset.from_hf_dataset(
        "test_memoryset_inverted_labels",
        hf_dataset.map(lambda x: {"label": 1 if x["label"] == 0 else 0}),
        value_column="text",
        embedding_model=PretrainedEmbeddingModel.GTE_BASE,
    )
    with model.use_memoryset(inverted_labeled_memoryset):
        predictions = model.predict(["Do you love soup?", "Are cats cute?"])
        assert predictions[0].label == 1
        assert predictions[1].label == 0

    predictions = model.predict(["Do you love soup?", "Are cats cute?"])
    assert predictions[0].label == 0
    assert predictions[1].label == 1


def test_predict_with_expected_labels(model: ClassificationModel):
    prediction = model.predict("Do you love soup?", expected_labels=1)
    assert prediction.expected_label == 1


def test_predict_with_expected_labels_invalid_input(model: ClassificationModel):
    # invalid number of expected labels for batch prediction
    with pytest.raises(ValueError, match=r"Invalid input.*"):
        model.predict(["Do you love soup?", "Are cats cute?"], expected_labels=[0])
    # invalid label value
    with pytest.raises(ValueError):
        model.predict("Do you love soup?", expected_labels=5)


def test_last_prediction_with_batch(model: ClassificationModel):
    predictions = model.predict(["Do you love soup?", "Are cats cute?"])
    assert model.last_prediction is not None
    assert model.last_prediction.prediction_id == predictions[-1].prediction_id
    assert model.last_prediction.input_value == "Are cats cute?"
    assert model._last_prediction_was_batch is True


def test_last_prediction_with_single(model: ClassificationModel):
    # Test that last_prediction is updated correctly with single prediction
    prediction = model.predict("Do you love soup?")
    assert model.last_prediction is not None
    assert model.last_prediction.prediction_id == prediction.prediction_id
    assert model.last_prediction.input_value == "Do you love soup?"
    assert model._last_prediction_was_batch is False



"""Contains methods for accessing the API"""

from .auth.check_authentication_auth_get import sync as check_authentication
from .auth.create_api_key_auth_api_key_post import sync as create_api_key
from .auth.delete_api_key_auth_api_key_name_or_id_delete import sync as delete_api_key
from .auth.delete_org_auth_org_delete import sync as delete_org
from .auth.list_api_keys_auth_api_key_get import sync as list_api_keys
from .classification_model.create_evaluation_classification_model_model_name_or_id_evaluation_post import (
    sync as create_evaluation,
)
from .classification_model.create_model_classification_model_post import (
    sync as create_model,
)
from .classification_model.delete_evaluation_classification_model_model_name_or_id_evaluation_task_id_delete import (
    sync as delete_evaluation,
)
from .classification_model.delete_model_classification_model_name_or_id_delete import (
    sync as delete_model,
)
from .classification_model.get_evaluation_classification_model_model_name_or_id_evaluation_task_id_get import (
    sync as get_evaluation,
)
from .classification_model.get_model_classification_model_name_or_id_get import (
    sync as get_model,
)
from .classification_model.list_evaluations_classification_model_model_name_or_id_evaluation_get import (
    sync as list_evaluations,
)
from .classification_model.list_models_classification_model_get import (
    sync as list_models,
)
from .classification_model.predict_gpu_classification_model_name_or_id_prediction_post import (
    sync as predict_gpu,
)
from .datasource.create_datasource_datasource_post import sync as create_datasource
from .datasource.create_embedding_evaluation_datasource_name_or_id_embedding_evaluation_post import (
    sync as create_embedding_evaluation,
)
from .datasource.delete_datasource_datasource_name_or_id_delete import (
    sync as delete_datasource,
)
from .datasource.get_datasource_datasource_name_or_id_get import sync as get_datasource
from .datasource.get_embedding_evaluation_datasource_name_or_id_embedding_evaluation_task_id_get import (
    sync as get_embedding_evaluation,
)
from .datasource.list_datasources_datasource_get import sync as list_datasources
from .datasource.list_embedding_evaluations_datasource_name_or_id_embedding_evaluation_get import (
    sync as list_embedding_evaluations,
)
from .default.healthcheck_get import sync as healthcheck
from .default.healthcheck_gpu_get import sync as healthcheck_gpu
from .finetuned_embedding_model.create_finetuned_embedding_model_finetuned_embedding_model_post import (
    sync as create_finetuned_embedding_model,
)
from .finetuned_embedding_model.delete_finetuned_embedding_model_finetuned_embedding_model_name_or_id_delete import (
    sync as delete_finetuned_embedding_model,
)
from .finetuned_embedding_model.embed_with_finetuned_model_gpu_finetuned_embedding_model_name_or_id_embedding_post import (
    sync as embed_with_finetuned_model_gpu,
)
from .finetuned_embedding_model.get_finetuned_embedding_model_finetuned_embedding_model_name_or_id_get import (
    sync as get_finetuned_embedding_model,
)
from .finetuned_embedding_model.list_finetuned_embedding_models_finetuned_embedding_model_get import (
    sync as list_finetuned_embedding_models,
)
from .memoryset.analyze_memoryset_memoryset_name_or_id_analysis_post import (
    sync as analyze_memoryset,
)
from .memoryset.clone_memoryset_memoryset_name_or_id_clone_post import (
    sync as clone_memoryset,
)
from .memoryset.create_memoryset_memoryset_post import sync as create_memoryset
from .memoryset.delete_memories_memoryset_name_or_id_memories_delete_post import (
    sync as delete_memories,
)
from .memoryset.delete_memory_memoryset_name_or_id_memory_memory_id_delete import (
    sync as delete_memory,
)
from .memoryset.delete_memoryset_memoryset_name_or_id_delete import (
    sync as delete_memoryset,
)
from .memoryset.get_analysis_memoryset_name_or_id_analysis_analysis_task_id_get import (
    sync as get_analysis,
)
from .memoryset.get_memories_memoryset_name_or_id_memories_get_post import (
    sync as get_memories,
)
from .memoryset.get_memory_memoryset_name_or_id_memory_memory_id_get import (
    sync as get_memory,
)
from .memoryset.get_memoryset_memoryset_name_or_id_get import sync as get_memoryset
from .memoryset.insert_memories_gpu_memoryset_name_or_id_memory_post import (
    sync as insert_memories_gpu,
)
from .memoryset.list_analyses_memoryset_name_or_id_analysis_get import (
    sync as list_analyses,
)
from .memoryset.list_memorysets_memoryset_get import sync as list_memorysets
from .memoryset.memoryset_lookup_gpu_memoryset_name_or_id_lookup_post import (
    sync as memoryset_lookup_gpu,
)
from .memoryset.query_memoryset_memoryset_name_or_id_memories_post import (
    sync as query_memoryset,
)
from .memoryset.update_memories_gpu_memoryset_name_or_id_memories_patch import (
    sync as update_memories_gpu,
)
from .memoryset.update_memory_gpu_memoryset_name_or_id_memory_patch import (
    sync as update_memory_gpu,
)
from .pretrained_embedding_model.embed_with_pretrained_model_gpu_pretrained_embedding_model_model_name_embedding_post import (
    sync as embed_with_pretrained_model_gpu,
)
from .pretrained_embedding_model.get_pretrained_embedding_model_pretrained_embedding_model_model_name_get import (
    sync as get_pretrained_embedding_model,
)
from .pretrained_embedding_model.list_pretrained_embedding_models_pretrained_embedding_model_get import (
    sync as list_pretrained_embedding_models,
)
from .task.abort_task_task_task_id_abort_delete import sync as abort_task
from .task.get_task_status_task_task_id_status_get import sync as get_task_status_task
from .task.list_tasks_task_get import sync as list_tasks
from .telemetry.count_predictions_telemetry_prediction_count_post import (
    sync as count_predictions,
)
from .telemetry.drop_feedback_category_with_data_telemetry_feedback_category_name_or_id_delete import (
    sync as drop_feedback_category_with_data,
)
from .telemetry.explain_prediction_telemetry_prediction_prediction_id_explanation_get import (
    sync as explain_prediction,
)
from .telemetry.get_feedback_category_telemetry_feedback_category_name_or_id_get import (
    sync as get_feedback_category,
)
from .telemetry.get_prediction_telemetry_prediction_prediction_id_get import (
    sync as get_prediction,
)
from .telemetry.list_feedback_categories_telemetry_feedback_category_get import (
    sync as list_feedback_categories,
)
from .telemetry.list_memories_with_feedback_telemetry_memories_post import (
    sync as list_memories_with_feedback,
)
from .telemetry.list_predictions_telemetry_prediction_post import (
    sync as list_predictions,
)
from .telemetry.record_prediction_feedback_telemetry_prediction_feedback_put import (
    sync as record_prediction_feedback,
)
from .telemetry.update_prediction_telemetry_prediction_prediction_id_patch import (
    sync as update_prediction,
)

__all__ = [
    "abort_task",
    "analyze_memoryset",
    "check_authentication",
    "clone_memoryset",
    "count_predictions",
    "create_api_key",
    "create_datasource",
    "create_embedding_evaluation",
    "create_evaluation",
    "create_finetuned_embedding_model",
    "create_memoryset",
    "create_model",
    "delete_api_key",
    "delete_datasource",
    "delete_evaluation",
    "delete_finetuned_embedding_model",
    "delete_memories",
    "delete_memory",
    "delete_memoryset",
    "delete_model",
    "delete_org",
    "drop_feedback_category_with_data",
    "embed_with_finetuned_model_gpu",
    "embed_with_pretrained_model_gpu",
    "explain_prediction",
    "get_analysis",
    "get_datasource",
    "get_embedding_evaluation",
    "get_evaluation",
    "get_feedback_category",
    "get_finetuned_embedding_model",
    "get_memories",
    "get_memory",
    "get_memoryset",
    "get_model",
    "get_prediction",
    "get_pretrained_embedding_model",
    "get_task_status_task",
    "healthcheck",
    "healthcheck_gpu",
    "insert_memories_gpu",
    "list_analyses",
    "list_api_keys",
    "list_datasources",
    "list_embedding_evaluations",
    "list_evaluations",
    "list_feedback_categories",
    "list_finetuned_embedding_models",
    "list_memories_with_feedback",
    "list_memorysets",
    "list_models",
    "list_predictions",
    "list_pretrained_embedding_models",
    "list_tasks",
    "memoryset_lookup_gpu",
    "predict_gpu",
    "query_memoryset",
    "record_prediction_feedback",
    "update_memories_gpu",
    "update_memory_gpu",
    "update_prediction",
]



"""
This file is generated by the openapi-python-client tool via the generate_api_client.py script

It is a customized template from the openapi-python-client tool's default template:
https://github.com/openapi-generators/openapi-python-client/blob/861ef5622f10fc96d240dc9becb0edf94e61446c/openapi_python_client/templates/model.py.jinja

The main change is:
- Fix typing issues
"""

# flake8: noqa: C901

from typing import TYPE_CHECKING, Any, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.memoryset_neighbor_metrics_lookup_score_metrics import MemorysetNeighborMetricsLookupScoreMetrics


T = TypeVar("T", bound="MemorysetNeighborMetrics")


@_attrs_define
class MemorysetNeighborMetrics:
    """
    Attributes:
        lookup_score_metrics (MemorysetNeighborMetricsLookupScoreMetrics):
    """

    lookup_score_metrics: "MemorysetNeighborMetricsLookupScoreMetrics"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        lookup_score_metrics = self.lookup_score_metrics.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "lookup_score_metrics": lookup_score_metrics,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: dict[str, Any]) -> T:
        from ..models.memoryset_neighbor_metrics_lookup_score_metrics import MemorysetNeighborMetricsLookupScoreMetrics

        d = src_dict.copy()
        lookup_score_metrics = MemorysetNeighborMetricsLookupScoreMetrics.from_dict(d.pop("lookup_score_metrics"))

        memoryset_neighbor_metrics = cls(
            lookup_score_metrics=lookup_score_metrics,
        )

        memoryset_neighbor_metrics.additional_properties = d
        return memoryset_neighbor_metrics

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class FilterItemFieldType0Item(str, Enum):
    CREATED_AT = "created_at"
    LABEL = "label"
    MEMORY_ID = "memory_id"
    METADATA = "metadata"
    METRICS = "metrics"
    SCORE = "score"
    SOURCE_ID = "source_id"
    UPDATED_AT = "updated_at"
    VALUE = "value"

    def __str__(self) -> str:
        return str(self.value)



from datetime import datetime
from typing import Literal, NamedTuple

from ._generated_api_client.api import (
    check_authentication,
    create_api_key,
    delete_api_key,
    list_api_keys,
)
from ._generated_api_client.client import get_base_url, get_headers, set_headers
from ._generated_api_client.models import (
    CreateApiKeyRequest,
    CreateApiKeyRequestScopeItem,
)

Scope = Literal["ADMINISTER", "PREDICT"]
"""
The scopes of an API key.

- `ADMINISTER`: Can do anything, including creating and deleting organizations, models, and API keys.
- `PREDICT`: Can only call model.predict and perform CRUD operations on predictions.
"""


class ApiKeyInfo(NamedTuple):
    """
    Named tuple containing information about an API key

    Attributes:
        name: Unique name of the API key
        created_at: When the API key was created
    """

    name: str
    created_at: datetime
    scopes: set[Scope]


class OrcaCredentials:
    """
    Class for managing Orca API credentials
    """

    @staticmethod
    def get_api_url() -> str:
        """
        Get the Orca API base URL that is currently being used
        """
        return get_base_url()

    @staticmethod
    def list_api_keys() -> list[ApiKeyInfo]:
        """
        List all API keys that have been created for your org

        Returns:
            A list of named tuples, with the name and creation date time of the API key
        """
        return [
            ApiKeyInfo(name=api_key.name, created_at=api_key.created_at, scopes=set(s.value for s in api_key.scope))
            for api_key in list_api_keys()
        ]

    @staticmethod
    def is_authenticated() -> bool:
        """
        Check if you are authenticated to interact with the Orca API

        Returns:
            True if you are authenticated, False otherwise
        """
        try:
            return check_authentication()
        except ValueError as e:
            if "Invalid API key" in str(e):
                return False
            raise e

    @staticmethod
    def create_api_key(name: str, scopes: set[Scope] = {"ADMINISTER"}) -> str:
        """
        Create a new API key with the given name and scopes

        Params:
            name: The name of the API key
            scopes: The scopes of the API key

        Returns:
            The secret value of the API key. Make sure to save this value as it will not be shown again.
        """
        res = create_api_key(
            body=CreateApiKeyRequest(name=name, scope=[CreateApiKeyRequestScopeItem(scope) for scope in scopes])
        )
        return res.api_key

    @staticmethod
    def revoke_api_key(name: str) -> None:
        """
        Delete an API key

        Params:
            name: The name of the API key to delete

        Raises:
            ValueError: if the API key is not found
        """
        delete_api_key(name_or_id=name)

    @staticmethod
    def set_api_key(api_key: str, check_validity: bool = True):
        """
        Set the API key to use for authenticating with the Orca API

        Note:
            The API key can also be provided by setting the `ORCA_API_KEY` environment variable

        Params:
            api_key: The API key to set
            check_validity: Whether to check if the API key is valid and raise an error otherwise

        Raises:
            ValueError: if the API key is invalid and `check_validity` is True
        """
        set_headers(get_headers() | {"Api-Key": api_key})
        if check_validity:
            check_authentication()



"""
This file is generated by the openapi-python-client tool via the generate_api_client.py script

It is a customized template from the openapi-python-client tool's default template:
https://github.com/openapi-generators/openapi-python-client/blob/861ef5622f10fc96d240dc9becb0edf94e61446c/openapi_python_client/templates/model.py.jinja

The main change is:
- Fix typing issues
"""

# flake8: noqa: C901

from enum import Enum
from typing import Any, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.prediction_sort_item_item_type_0 import PredictionSortItemItemType0
from ..models.prediction_sort_item_item_type_1 import PredictionSortItemItemType1
from ..types import UNSET, Unset

T = TypeVar("T", bound="ListPredictionsRequest")


@_attrs_define
class ListPredictionsRequest:
    """
    Attributes:
        model_id (Union[None, Unset, str]):
        tag (Union[None, Unset, str]):
        prediction_ids (Union[List[str], None, Unset]):
        limit (Union[None, Unset, int]):
        offset (Union[None, Unset, int]):  Default: 0.
        sort (Union[Unset, List[List[Union[PredictionSortItemItemType0, PredictionSortItemItemType1]]]]):
        expected_label_match (Union[None, Unset, bool]):
    """

    model_id: Union[None, Unset, str] = UNSET
    tag: Union[None, Unset, str] = UNSET
    prediction_ids: Union[List[str], None, Unset] = UNSET
    limit: Union[None, Unset, int] = UNSET
    offset: Union[None, Unset, int] = 0
    sort: Union[Unset, List[List[Union[PredictionSortItemItemType0, PredictionSortItemItemType1]]]] = UNSET
    expected_label_match: Union[None, Unset, bool] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        model_id: Union[None, Unset, str]
        if isinstance(self.model_id, Unset):
            model_id = UNSET
        else:
            model_id = self.model_id

        tag: Union[None, Unset, str]
        if isinstance(self.tag, Unset):
            tag = UNSET
        else:
            tag = self.tag

        prediction_ids: Union[List[str], None, Unset]
        if isinstance(self.prediction_ids, Unset):
            prediction_ids = UNSET
        elif isinstance(self.prediction_ids, list):
            prediction_ids = self.prediction_ids

        else:
            prediction_ids = self.prediction_ids

        limit: Union[None, Unset, int]
        if isinstance(self.limit, Unset):
            limit = UNSET
        else:
            limit = self.limit

        offset: Union[None, Unset, int]
        if isinstance(self.offset, Unset):
            offset = UNSET
        else:
            offset = self.offset

        sort: Union[Unset, List[List[str]]] = UNSET
        if not isinstance(self.sort, Unset):
            sort = []
            for componentsschemas_prediction_sort_item_data in self.sort:
                componentsschemas_prediction_sort_item = []
                for componentsschemas_prediction_sort_item_item_data in componentsschemas_prediction_sort_item_data:
                    componentsschemas_prediction_sort_item_item: str
                    if isinstance(componentsschemas_prediction_sort_item_item_data, PredictionSortItemItemType0):
                        componentsschemas_prediction_sort_item_item = (
                            componentsschemas_prediction_sort_item_item_data.value
                            if isinstance(componentsschemas_prediction_sort_item_item_data, Enum)
                            else componentsschemas_prediction_sort_item_item_data
                        )
                    else:
                        componentsschemas_prediction_sort_item_item = (
                            componentsschemas_prediction_sort_item_item_data.value
                            if isinstance(componentsschemas_prediction_sort_item_item_data, Enum)
                            else componentsschemas_prediction_sort_item_item_data
                        )

                    componentsschemas_prediction_sort_item.append(componentsschemas_prediction_sort_item_item)

                sort.append(componentsschemas_prediction_sort_item)

        expected_label_match: Union[None, Unset, bool]
        if isinstance(self.expected_label_match, Unset):
            expected_label_match = UNSET
        else:
            expected_label_match = self.expected_label_match

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if model_id is not UNSET:
            field_dict["model_id"] = model_id
        if tag is not UNSET:
            field_dict["tag"] = tag
        if prediction_ids is not UNSET:
            field_dict["prediction_ids"] = prediction_ids
        if limit is not UNSET:
            field_dict["limit"] = limit
        if offset is not UNSET:
            field_dict["offset"] = offset
        if sort is not UNSET:
            field_dict["sort"] = sort
        if expected_label_match is not UNSET:
            field_dict["expected_label_match"] = expected_label_match

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()

        def _parse_model_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        model_id = _parse_model_id(d.pop("model_id", UNSET))

        def _parse_tag(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        tag = _parse_tag(d.pop("tag", UNSET))

        def _parse_prediction_ids(data: object) -> Union[List[str], None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                prediction_ids_type_0 = cast(List[str], data)

                return prediction_ids_type_0
            except:  # noqa: E722
                pass
            return cast(Union[List[str], None, Unset], data)

        prediction_ids = _parse_prediction_ids(d.pop("prediction_ids", UNSET))

        def _parse_limit(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        limit = _parse_limit(d.pop("limit", UNSET))

        def _parse_offset(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        offset = _parse_offset(d.pop("offset", UNSET))

        sort = []
        _sort = d.pop("sort", UNSET)
        for componentsschemas_prediction_sort_item_data in _sort or []:
            componentsschemas_prediction_sort_item = []
            _componentsschemas_prediction_sort_item = componentsschemas_prediction_sort_item_data
            for componentsschemas_prediction_sort_item_item_data in _componentsschemas_prediction_sort_item:

                def _parse_componentsschemas_prediction_sort_item_item(
                    data: object,
                ) -> Union[PredictionSortItemItemType0, PredictionSortItemItemType1]:
                    try:
                        if not isinstance(data, str):
                            raise TypeError()
                        componentsschemas_prediction_sort_item_item_type_0 = PredictionSortItemItemType0(data)

                        return componentsschemas_prediction_sort_item_item_type_0
                    except:  # noqa: E722
                        pass
                    if not isinstance(data, str):
                        raise TypeError()
                    componentsschemas_prediction_sort_item_item_type_1 = PredictionSortItemItemType1(data)

                    return componentsschemas_prediction_sort_item_item_type_1

                componentsschemas_prediction_sort_item_item = _parse_componentsschemas_prediction_sort_item_item(
                    componentsschemas_prediction_sort_item_item_data
                )

                componentsschemas_prediction_sort_item.append(componentsschemas_prediction_sort_item_item)

            sort.append(componentsschemas_prediction_sort_item)

        def _parse_expected_label_match(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        expected_label_match = _parse_expected_label_match(d.pop("expected_label_match", UNSET))

        list_predictions_request = cls(
            model_id=model_id,
            tag=tag,
            prediction_ids=prediction_ids,
            limit=limit,
            offset=offset,
            sort=sort,
            expected_label_match=expected_label_match,
        )

        list_predictions_request.additional_properties = d
        return list_predictions_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import logging
from contextlib import contextmanager
from datetime import datetime
from typing import Any, Generator, Iterable, Literal, cast, overload
from uuid import UUID

from ._generated_api_client.api import (
    create_evaluation,
    create_model,
    delete_model,
    get_evaluation,
    get_model,
    list_models,
    list_predictions,
    predict_gpu,
    record_prediction_feedback,
)
from ._generated_api_client.models import (
    CreateRACModelRequest,
    EvaluationRequest,
    ListPredictionsRequest,
)
from ._generated_api_client.models import (
    PredictionSortItemItemType0 as PredictionSortColumns,
)
from ._generated_api_client.models import (
    PredictionSortItemItemType1 as PredictionSortDirection,
)
from ._generated_api_client.models import (
    RACHeadType,
    RACModelMetadata,
)
from ._generated_api_client.models.prediction_request import PredictionRequest
from ._utils.common import CreateMode, DropMode
from ._utils.task import wait_for_task
from .datasource import Datasource
from .memoryset import LabeledMemoryset
from .telemetry import LabelPrediction, _parse_feedback


class ClassificationModel:
    """
    A handle to a classification model in OrcaCloud

    Attributes:
        id: Unique identifier for the model
        name: Unique name of the model
        description: Optional description of the model
        memoryset: Memoryset that the model uses
        head_type: Classification head type of the model
        num_classes: Number of distinct classes the model can predict
        memory_lookup_count: Number of memories the model uses for each prediction
        weigh_memories: If using a KNN head, whether the model weighs memories by their lookup score
        min_memory_weight: If using a KNN head, minimum lookup score memories have to be over to not be ignored
        created_at: When the model was created
    """

    id: str
    name: str
    description: str | None
    memoryset: LabeledMemoryset
    head_type: RACHeadType
    num_classes: int
    memory_lookup_count: int
    weigh_memories: bool | None
    min_memory_weight: float | None
    version: int
    created_at: datetime

    def __init__(self, metadata: RACModelMetadata):
        # for internal use only, do not document
        self.id = metadata.id
        self.name = metadata.name
        self.description = metadata.description
        self.memoryset = LabeledMemoryset.open(metadata.memoryset_id)
        self.head_type = metadata.head_type
        self.num_classes = metadata.num_classes
        self.memory_lookup_count = metadata.memory_lookup_count
        self.weigh_memories = metadata.weigh_memories
        self.min_memory_weight = metadata.min_memory_weight
        self.version = metadata.version
        self.created_at = metadata.created_at

        self._memoryset_override_id: str | None = None
        self._last_prediction: LabelPrediction | None = None
        self._last_prediction_was_batch: bool = False

    def __eq__(self, other) -> bool:
        return isinstance(other, ClassificationModel) and self.id == other.id

    def __repr__(self):
        return (
            "ClassificationModel({\n"
            f"    name: '{self.name}',\n"
            f"    head_type: {self.head_type},\n"
            f"    num_classes: {self.num_classes},\n"
            f"    memory_lookup_count: {self.memory_lookup_count},\n"
            f"    memoryset: LabeledMemoryset.open('{self.memoryset.name}'),\n"
            "})"
        )

    @property
    def last_prediction(self) -> LabelPrediction:
        """
        Last prediction made by the model

        Note:
            If the last prediction was part of a batch prediction, the last prediction from the
            batch is returned. If no prediction has been made yet, a [`LookupError`][LookupError]
            is raised.
        """
        if self._last_prediction_was_batch:
            logging.warning(
                "Last prediction was part of a batch prediction, returning the last prediction from the batch"
            )
        if self._last_prediction is None:
            raise LookupError("No prediction has been made yet")
        return self._last_prediction

    @classmethod
    def create(
        cls,
        name: str,
        memoryset: LabeledMemoryset,
        head_type: Literal["BMMOE", "FF", "KNN", "MMOE"] = "KNN",
        *,
        description: str | None = None,
        num_classes: int | None = None,
        memory_lookup_count: int | None = None,
        weigh_memories: bool = True,
        min_memory_weight: float | None = None,
        if_exists: CreateMode = "error",
    ) -> ClassificationModel:
        """
        Create a new classification model

        Params:
            name: Name for the new model (must be unique)
            memoryset: Memoryset to attach the model to
            head_type: Type of model head to use
            num_classes: Number of classes this model can predict, will be inferred from memoryset if not specified
            memory_lookup_count: Number of memories to lookup for each prediction,
                by default the system uses a simple heuristic to choose a number of memories that works well in most cases
            weigh_memories: If using a KNN head, whether the model weighs memories by their lookup score
            min_memory_weight: If using a KNN head, minimum lookup score memories have to be over to not be ignored
            if_exists: What to do if a model with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing model.
            description: Optional description for the model, this will be used in agentic flows,
                so make sure it is concise and describes the purpose of your model.

        Returns:
            Handle to the new model in the OrcaCloud

        Raises:
            ValueError: If the model already exists and if_exists is `"error"` or if it is
                `"open"` and the existing model has different attributes.

        Examples:
            Create a new model using default options:
            >>> model = ClassificationModel.create(
            ...    "my_model",
            ...    LabeledMemoryset.open("my_memoryset"),
            ... )

            Create a new model with non-default model head and options:
            >>> model = ClassificationModel.create(
            ...     name="my_model",
            ...     memoryset=LabeledMemoryset.open("my_memoryset"),
            ...     head_type=RACHeadType.MMOE,
            ...     num_classes=5,
            ...     memory_lookup_count=20,
            ... )
        """
        if cls.exists(name):
            if if_exists == "error":
                raise ValueError(f"Model with name {name} already exists")
            elif if_exists == "open":
                existing = cls.open(name)
                for attribute in {"head_type", "memory_lookup_count", "num_classes", "min_memory_weight"}:
                    local_attribute = locals()[attribute]
                    existing_attribute = getattr(existing, attribute)
                    if local_attribute is not None and local_attribute != existing_attribute:
                        raise ValueError(f"Model with name {name} already exists with different {attribute}")

                # special case for memoryset
                if existing.memoryset.id != memoryset.id:
                    raise ValueError(f"Model with name {name} already exists with different memoryset")

                return existing

        metadata = create_model(
            body=CreateRACModelRequest(
                name=name,
                memoryset_id=memoryset.id,
                head_type=RACHeadType(head_type),
                memory_lookup_count=memory_lookup_count,
                num_classes=num_classes,
                weigh_memories=weigh_memories,
                min_memory_weight=min_memory_weight,
                description=description,
            ),
        )
        return cls(metadata)

    @classmethod
    def open(cls, name: str) -> ClassificationModel:
        """
        Get a handle to a classification model in the OrcaCloud

        Params:
            name: Name or unique identifier of the classification model

        Returns:
            Handle to the existing classification model in the OrcaCloud

        Raises:
            LookupError: If the classification model does not exist
        """
        return cls(get_model(name))

    @classmethod
    def exists(cls, name_or_id: str) -> bool:
        """
        Check if a classification model exists in the OrcaCloud

        Params:
            name_or_id: Name or id of the classification model

        Returns:
            `True` if the classification model exists, `False` otherwise
        """
        try:
            cls.open(name_or_id)
            return True
        except LookupError:
            return False

    @classmethod
    def all(cls) -> list[ClassificationModel]:
        """
        Get a list of handles to all classification models in the OrcaCloud

        Returns:
            List of handles to all classification models in the OrcaCloud
        """
        return [cls(metadata) for metadata in list_models()]

    @classmethod
    def drop(cls, name_or_id: str, if_not_exists: DropMode = "error"):
        """
        Delete a classification model from the OrcaCloud

        Warning:
            This will delete the model and all associated data, including predictions, evaluations, and feedback.

        Params:
            name_or_id: Name or id of the classification model
            if_not_exists: What to do if the classification model does not exist, defaults to `"error"`.
                Other option is `"ignore"` to do nothing if the classification model does not exist.

        Raises:
            LookupError: If the classification model does not exist and if_not_exists is `"error"`
        """
        try:
            delete_model(name_or_id)
            logging.info(f"Deleted model {name_or_id}")
        except LookupError:
            if if_not_exists == "error":
                raise

    @overload
    def predict(
        self, value: list[str], expected_labels: list[int] | None = None, tags: set[str] = set()
    ) -> list[LabelPrediction]:
        pass

    @overload
    def predict(self, value: str, expected_labels: int | None = None, tags: set[str] = set()) -> LabelPrediction:
        pass

    def predict(
        self, value: list[str] | str, expected_labels: list[int] | int | None = None, tags: set[str] = set()
    ) -> list[LabelPrediction] | LabelPrediction:
        """
        Predict label(s) for the given input value(s) grounded in similar memories

        Params:
            value: Value(s) to get predict the labels of
            expected_labels: Expected label(s) for the given input to record for model evaluation
            tags: Tags to add to the prediction(s)

        Returns:
            Label prediction or list of label predictions

        Examples:
            Predict the label for a single value:
            >>> prediction = model.predict("I am happy", tags={"test"})
            LabelPrediction({label: <positive: 1>, confidence: 0.95, anomaly_score: 0.1, input_value: 'I am happy' })

            Predict the labels for a list of values:
            >>> predictions = model.predict(["I am happy", "I am sad"], expected_labels=[1, 0])
            [
                LabelPrediction({label: <positive: 1>, confidence: 0.95, anomaly_score: 0.1, input_value: 'I am happy'}),
                LabelPrediction({label: <negative: 0>, confidence: 0.05, anomaly_score: 0.1, input_value: 'I am sad'}),
            ]
        """
        response = predict_gpu(
            self.id,
            body=PredictionRequest(
                input_values=value if isinstance(value, list) else [value],
                memoryset_override_id=self._memoryset_override_id,
                expected_labels=(
                    expected_labels
                    if isinstance(expected_labels, list)
                    else [expected_labels] if expected_labels is not None else None
                ),
                tags=list(tags),
            ),
        )
        predictions = [
            LabelPrediction(
                prediction_id=prediction.prediction_id,
                label=prediction.label,
                label_name=prediction.label_name,
                confidence=prediction.confidence,
                anomaly_score=prediction.anomaly_score,
                memoryset=self.memoryset,
                model=self,
            )
            for prediction in response
        ]
        self._last_prediction_was_batch = isinstance(value, list)
        self._last_prediction = predictions[-1]
        return predictions if isinstance(value, list) else predictions[0]

    def predictions(
        self,
        limit: int = 100,
        offset: int = 0,
        tag: str | None = None,
        sort: list[tuple[PredictionSortColumns, PredictionSortDirection]] = [],
        expected_label_match: bool | None = None,
    ) -> list[LabelPrediction]:
        """
        Get a list of predictions made by this model

        Params:
            limit: Optional maximum number of predictions to return
            offset: Optional offset of the first prediction to return
            tag: Optional tag to filter predictions by
            sort: Optional list of columns and directions to sort the predictions by.
                Predictions can be sorted by `timestamp` or `confidence`.
            expected_label_match: Optional filter to only include predictions where the expected
                label does (`True`) or doesn't (`False`) match the predicted label

        Returns:
            List of label predictions

        Examples:
            Get the last 3 predictions:
            >>> predictions = model.predictions(limit=3, sort=[("timestamp", "desc")])
            [
                LabeledPrediction({label: <positive: 1>, confidence: 0.95, anomaly_score: 0.1, input_value: 'I am happy'}),
                LabeledPrediction({label: <negative: 0>, confidence: 0.05, anomaly_score: 0.1, input_value: 'I am sad'}),
                LabeledPrediction({label: <positive: 1>, confidence: 0.90, anomaly_score: 0.1, input_value: 'I am ecstatic'}),
            ]


            Get second most confident prediction:
            >>> predictions = model.predictions(sort=[("confidence", "desc")], offset=1, limit=1)
            [LabeledPrediction({label: <positive: 1>, confidence: 0.90, anomaly_score: 0.1, input_value: 'I am having a good day'})]

            Get predictions where the expected label doesn't match the predicted label:
            >>> predictions = model.predictions(expected_label_match=False)
            [LabeledPrediction({label: <positive: 1>, confidence: 0.95, anomaly_score: 0.1, input_value: 'I am happy', expected_label: 0})]
        """
        predictions = list_predictions(
            body=ListPredictionsRequest(
                model_id=self.id,
                limit=limit,
                offset=offset,
                sort=cast(list[list[PredictionSortColumns | PredictionSortDirection]], sort),
                tag=tag,
                expected_label_match=expected_label_match,
            ),
        )
        return [
            LabelPrediction(
                prediction_id=prediction.prediction_id,
                label=prediction.label,
                label_name=prediction.label_name,
                confidence=prediction.confidence,
                anomaly_score=prediction.anomaly_score,
                memoryset=self.memoryset,
                model=self,
                telemetry=prediction,
            )
            for prediction in predictions
        ]

    def evaluate(
        self,
        datasource: Datasource,
        value_column: str = "value",
        label_column: str = "label",
        record_predictions: bool = False,
        tags: set[str] | None = None,
    ) -> dict[str, Any]:
        """
        Evaluate the classification model on a given datasource

        Params:
            datasource: Datasource to evaluate the model on
            value_column: Name of the column that contains the input values to the model
            label_column: Name of the column containing the expected labels
            record_predictions: Whether to record [`LabelPrediction`][orca_sdk.telemetry.LabelPrediction]s for analysis
            tags: Optional tags to add to the recorded [`LabelPrediction`][orca_sdk.telemetry.LabelPrediction]s

        Returns:
            Dictionary with evaluation metrics

        Examples:
            >>> model.evaluate(datasource, value_column="text", label_column="airline_sentiment")
            { "f1_score": 0.85, "roc_auc": 0.85, "pr_auc": 0.85, "accuracy": 0.85, "loss": 0.35, ... }
        """
        response = create_evaluation(
            self.id,
            body=EvaluationRequest(
                datasource_id=datasource.id,
                datasource_label_column=label_column,
                datasource_value_column=value_column,
                memoryset_override_id=self._memoryset_override_id,
                record_telemetry=record_predictions,
                telemetry_tags=list(tags) if tags else None,
            ),
        )
        wait_for_task(response.task_id, description="Running evaluation")
        response = get_evaluation(self.id, UUID(response.task_id))
        assert response.result is not None
        return response.result.to_dict()

    def finetune(self, datasource: Datasource):
        #  do not document until implemented
        raise NotImplementedError("Finetuning is not supported yet")

    @contextmanager
    def use_memoryset(self, memoryset_override: LabeledMemoryset) -> Generator[None, None, None]:
        """
        Temporarily override the memoryset used by the model for predictions

        Params:
            memoryset_override: Memoryset to override the default memoryset with

        Examples:
            >>> with model.use_memoryset(LabeledMemoryset.open("my_other_memoryset")):
            ...     predictions = model.predict("I am happy")
        """
        self._memoryset_override_id = memoryset_override.id
        yield
        self._memoryset_override_id = None

    @overload
    def record_feedback(self, feedback: dict[str, Any]) -> None:
        pass

    @overload
    def record_feedback(self, feedback: Iterable[dict[str, Any]]) -> None:
        pass

    def record_feedback(self, feedback: Iterable[dict[str, Any]] | dict[str, Any]):
        """
        Record feedback for a list of predictions.

        We support recording feedback in several categories for each prediction. A
        [`FeedbackCategory`][orca_sdk.telemetry.FeedbackCategory] is created automatically,
        the first time feedback with a new name is recorded. Categories are global across models.
        The value type of the category is inferred from the first recorded value. Subsequent
        feedback for the same category must be of the same type.

        Params:
            feedback: Feedback to record, this should be dictionaries with the following keys:

                - `category`: Name of the category under which to record the feedback.
                - `value`: Feedback value to record, should be `True` for positive feedback and
                    `False` for negative feedback or a [`float`][float] between `-1.0` and `+1.0`
                    where negative values indicate negative feedback and positive values indicate
                    positive feedback.
                - `comment`: Optional comment to record with the feedback.

        Examples:
            Record whether predictions were correct or incorrect:
            >>> model.record_feedback({
            ...     "prediction": p.prediction_id,
            ...     "category": "correct",
            ...     "value": p.label == p.expected_label,
            ... } for p in predictions)

            Record star rating as normalized continuous score between `-1.0` and `+1.0`:
            >>> model.record_feedback({
            ...     "prediction": "123e4567-e89b-12d3-a456-426614174000",
            ...     "category": "rating",
            ...     "value": -0.5,
            ...     "comment": "2 stars"
            ... })

        Raises:
            ValueError: If the value does not match previous value types for the category, or is a
                [`float`][float] that is not between `-1.0` and `+1.0`.
        """
        record_prediction_feedback(
            body=[
                _parse_feedback(f) for f in (cast(list[dict], [feedback]) if isinstance(feedback, dict) else feedback)
            ],
        )



"""
This file is generated by the openapi-python-client tool via the generate_api_client.py script

It is a customized template from the openapi-python-client tool's default template:
https://github.com/openapi-generators/openapi-python-client/blob/861ef5622f10fc96d240dc9becb0edf94e61446c/openapi_python_client/templates/model.py.jinja

The main change is:
- Fix typing issues
"""

# flake8: noqa: C901

from typing import Any, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="PredictionRequest")


@_attrs_define
class PredictionRequest:
    """
    Attributes:
        input_values (List[str]):
        expected_labels (Union[List[int], None, Unset]):
        tags (Union[Unset, List[str]]):
        memoryset_override_id (Union[None, Unset, str]):
    """

    input_values: List[str]
    expected_labels: Union[List[int], None, Unset] = UNSET
    tags: Union[Unset, List[str]] = UNSET
    memoryset_override_id: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        input_values: List[str]
        if isinstance(self.input_values, list):
            input_values = self.input_values

        expected_labels: Union[List[int], None, Unset]
        if isinstance(self.expected_labels, Unset):
            expected_labels = UNSET
        elif isinstance(self.expected_labels, list):
            expected_labels = self.expected_labels

        else:
            expected_labels = self.expected_labels

        tags: Union[Unset, List[str]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags

        memoryset_override_id: Union[None, Unset, str]
        if isinstance(self.memoryset_override_id, Unset):
            memoryset_override_id = UNSET
        else:
            memoryset_override_id = self.memoryset_override_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "input_values": input_values,
            }
        )
        if expected_labels is not UNSET:
            field_dict["expected_labels"] = expected_labels
        if tags is not UNSET:
            field_dict["tags"] = tags
        if memoryset_override_id is not UNSET:
            field_dict["memoryset_override_id"] = memoryset_override_id

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()

        def _parse_input_values(data: object) -> List[str]:
            try:
                if not isinstance(data, list):
                    raise TypeError()
                input_values_type_0 = cast(List[str], data)

                return input_values_type_0
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, list):
                    raise TypeError()
                input_values_type_1 = cast(List[str], data)

                return input_values_type_1
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, list):
                    raise TypeError()
                input_values_type_2 = cast(List[str], data)

                return input_values_type_2
            except:  # noqa: E722
                pass
            if not isinstance(data, list):
                raise TypeError()
            input_values_type_3 = []
            _input_values_type_3 = data
            for input_values_type_3_item_data in _input_values_type_3:

                def _parse_input_values_type_3_item(data: object) -> str:
                    return cast(str, data)

                input_values_type_3_item = _parse_input_values_type_3_item(input_values_type_3_item_data)

                input_values_type_3.append(input_values_type_3_item)

            return input_values_type_3

        input_values = _parse_input_values(d.pop("input_values"))

        def _parse_expected_labels(data: object) -> Union[List[int], None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                expected_labels_type_0 = cast(List[int], data)

                return expected_labels_type_0
            except:  # noqa: E722
                pass
            return cast(Union[List[int], None, Unset], data)

        expected_labels = _parse_expected_labels(d.pop("expected_labels", UNSET))

        tags = cast(List[str], d.pop("tags", UNSET))

        def _parse_memoryset_override_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        memoryset_override_id = _parse_memoryset_override_id(d.pop("memoryset_override_id", UNSET))

        prediction_request = cls(
            input_values=input_values,
            expected_labels=expected_labels,
            tags=tags,
            memoryset_override_id=memoryset_override_id,
        )

        prediction_request.additional_properties = d
        return prediction_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from __future__ import annotations

import logging
from datetime import datetime
from typing import TYPE_CHECKING, Any, Generator, Iterable, overload
from uuid import UUID

from orca_sdk._utils.common import UNSET

from ._generated_api_client.api import (
    drop_feedback_category_with_data,
    get_prediction,
    list_feedback_categories,
    list_predictions,
    record_prediction_feedback,
    update_prediction,
)
from ._generated_api_client.client import get_client
from ._generated_api_client.errors import get_error_for_response
from ._generated_api_client.models import (
    FeedbackType,
    LabelPredictionWithMemoriesAndFeedback,
    ListPredictionsRequest,
    PredictionFeedbackCategory,
    PredictionFeedbackRequest,
    UpdatePredictionRequest,
)
from ._generated_api_client.types import UNSET as CLIENT_UNSET
from ._utils.prediction_result_ui import inspect_prediction_result
from .memoryset import LabeledMemoryLookup, LabeledMemoryset

if TYPE_CHECKING:
    from .classification_model import ClassificationModel


def _parse_feedback(feedback: dict[str, Any]) -> PredictionFeedbackRequest:
    category = feedback.get("category", None)
    if category is None:
        raise ValueError("`category` must be specified")
    prediction_id = feedback.get("prediction_id", None)
    if prediction_id is None:
        raise ValueError("`prediction_id` must be specified")
    return PredictionFeedbackRequest(
        prediction_id=prediction_id,
        category_name=category,
        value=feedback.get("value", CLIENT_UNSET),
        comment=feedback.get("comment", CLIENT_UNSET),
    )


class FeedbackCategory:
    """
    A category of feedback for predictions.

    Categories are created automatically, the first time feedback with a new name is recorded.
    The value type of the category is inferred from the first recorded value. Subsequent feedback
    for the same category must be of the same type. Categories are not model specific.

    Attributes:
        id: Unique identifier for the category.
        name: Name of the category.
        value_type: Type that values for this category must have.
        created_at: When the category was created.
    """

    id: str
    name: str
    value_type: type[bool] | type[float]
    created_at: datetime

    def __init__(self, category: PredictionFeedbackCategory):
        # for internal use only, do not document
        self.id = category.id
        self.name = category.name
        self.value_type = bool if category.type == FeedbackType.BINARY else float
        self.created_at = category.created_at

    @classmethod
    def all(cls) -> list[FeedbackCategory]:
        """
        Get a list of all existing feedback categories.

        Returns:
            List with information about all existing feedback categories.
        """
        return [FeedbackCategory(category) for category in list_feedback_categories()]

    @classmethod
    def drop(cls, name: str) -> None:
        """
        Drop all feedback for this category and drop the category itself, allowing it to be
        recreated with a different value type.

        Warning:
            This will delete all feedback in this category across all models.

        Params:
            name: Name of the category to drop.

        Raises:
            LookupError: If the category is not found.
        """
        drop_feedback_category_with_data(name)
        logging.info(f"Deleted feedback category {name} with all associated feedback")

    def __repr__(self):
        return "FeedbackCategory({" + f"name: {self.name}, " + f"value_type: {self.value_type}" + "})"


class LabelPrediction:
    """
    A prediction made by a model

    Attributes:
        prediction_id: Unique identifier for the prediction
        label: Predicted label for the input value
        label_name: Name of the predicted label
        confidence: Confidence of the prediction
        anomaly_score: The score for how anomalous the input is relative to the memories
        memory_lookups: List of memories used to ground the prediction
        input_value: Input value that this prediction was for
        model: Model that was used to make the prediction
        memoryset: Memoryset that was used to lookup memories to ground the prediction
        expected_label: Optional expected label that was set for the prediction
        expected_label_name: Name of the expected label
        tags: tags that were set for the prediction
        feedback: Feedback recorded, mapping from category name to value
        explanation: Explanation why the model made the prediction generated by a reasoning agent
    """

    prediction_id: str
    label: int
    label_name: str | None
    confidence: float
    anomaly_score: float | None
    memoryset: LabeledMemoryset
    model: ClassificationModel

    def __init__(
        self,
        prediction_id: str,
        *,
        label: int,
        label_name: str | None,
        confidence: float,
        anomaly_score: float | None,
        memoryset: LabeledMemoryset | str,
        model: ClassificationModel | str,
        telemetry: LabelPredictionWithMemoriesAndFeedback | None = None,
    ):
        # for internal use only, do not document
        from .classification_model import ClassificationModel

        self.prediction_id = prediction_id
        self.label = label
        self.label_name = label_name
        self.confidence = confidence
        self.anomaly_score = anomaly_score
        self.memoryset = LabeledMemoryset.open(memoryset) if isinstance(memoryset, str) else memoryset
        self.model = ClassificationModel.open(model) if isinstance(model, str) else model
        self.__telemetry = telemetry if telemetry else None

    def __repr__(self):
        return (
            "LabelPrediction({"
            + f"label: <{self.label_name}: {self.label}>, "
            + f"confidence: {self.confidence:.2f}, "
            + f"anomaly_score: {self.anomaly_score:.2f}, "
            if self.anomaly_score is not None
            else ""
            + f"input_value: '{str(self.input_value)[:100] + '...' if len(str(self.input_value)) > 100 else self.input_value}'"
            + "})"
        )

    @property
    def _telemetry(self) -> LabelPredictionWithMemoriesAndFeedback:
        # for internal use only, do not document
        if self.__telemetry is None:
            self.__telemetry = get_prediction(prediction_id=UUID(self.prediction_id))
        return self.__telemetry

    @property
    def memory_lookups(self) -> list[LabeledMemoryLookup]:
        return [LabeledMemoryLookup(self.memoryset.id, lookup) for lookup in self._telemetry.memories]

    @property
    def input_value(self) -> str | list[list[float]] | None:
        return self._telemetry.input_value

    @property
    def feedback(self) -> dict[str, bool | float]:
        return {
            f.category_name: (
                f.value if f.category_type == FeedbackType.CONTINUOUS else True if f.value == 1 else False
            )
            for f in self._telemetry.feedbacks
        }

    @property
    def expected_label(self) -> int | None:
        return self._telemetry.expected_label

    @property
    def expected_label_name(self) -> str | None:
        return self._telemetry.expected_label_name

    @property
    def tags(self) -> set[str]:
        return set(self._telemetry.tags)

    def _explanation_stream(self, refresh: bool = False) -> Generator[str, None, None]:
        httpx_client = get_client().get_httpx_client()
        url = f"/telemetry/prediction/{self.prediction_id}/explanation?refresh={refresh}"
        with httpx_client.stream("GET", url) as res:
            if res.status_code != 200:
                raise get_error_for_response(res)
            for chunk in res.iter_text():
                yield chunk

    @property
    def explanation(self) -> str:
        if self._telemetry.explanation is None:
            self._telemetry.explanation = "".join(self._explanation_stream())
        return self._telemetry.explanation

    def explain(self, refresh: bool = False) -> None:
        """
        Print an explanation of the prediction as a stream of text.

        Params:
            refresh: Force the explanation agent to re-run even if an explanation already exists.
        """
        if not refresh and self._telemetry.explanation is not None:
            print(self._telemetry.explanation)
        else:
            for chunk in self._explanation_stream(refresh):
                print(chunk, end="")
            print()  # final newline

    @overload
    @classmethod
    def get(cls, prediction_id: str) -> LabelPrediction:  # type: ignore -- this takes precedence
        pass

    @overload
    @classmethod
    def get(cls, prediction_id: Iterable[str]) -> list[LabelPrediction]:
        pass

    @classmethod
    def get(cls, prediction_id: str | Iterable[str]) -> LabelPrediction | list[LabelPrediction]:
        """
        Fetch a prediction or predictions

        Params:
            prediction_id: Unique identifier of the prediction or predictions to fetch

        Returns:
            Prediction or list of predictions

        Raises:
            LookupError: If no prediction with the given id is found

        Examples:
            Fetch a single prediction:
            >>> LabelPrediction.get("0195019a-5bc7-7afb-b902-5945ee1fb766")
            LabelPrediction({
                label: <positive: 1>,
                confidence: 0.95,
                anomaly_score: 0.1,
                input_value: "I am happy",
                memoryset: "my_memoryset",
                model: "my_model"
            })

            Fetch multiple predictions:
            >>> LabelPrediction.get([
            ...     "0195019a-5bc7-7afb-b902-5945ee1fb766",
            ...     "019501a1-ea08-76b2-9f62-95e4800b4841",
            ... ])
            [
                LabelPrediction({
                    label: <positive: 1>,
                    confidence: 0.95,
                    anomaly_score: 0.1,
                    input_value: "I am happy",
                    memoryset: "my_memoryset",
                    model: "my_model"
                }),
                LabelPrediction({
                    label: <negative: 0>,
                    confidence: 0.05,
                    anomaly_score: 0.2,
                    input_value: "I am sad",
                    memoryset: "my_memoryset", model: "my_model"
                }),
            ]
        """
        if isinstance(prediction_id, str):
            prediction = get_prediction(prediction_id=UUID(prediction_id))
            return cls(
                prediction_id=prediction.prediction_id,
                label=prediction.label,
                label_name=prediction.label_name,
                confidence=prediction.confidence,
                anomaly_score=prediction.anomaly_score,
                memoryset=prediction.memoryset_id,
                model=prediction.model_id,
                telemetry=prediction,
            )
        else:
            return [
                cls(
                    prediction_id=prediction.prediction_id,
                    label=prediction.label,
                    label_name=prediction.label_name,
                    confidence=prediction.confidence,
                    anomaly_score=prediction.anomaly_score,
                    memoryset=prediction.memoryset_id,
                    model=prediction.model_id,
                    telemetry=prediction,
                )
                for prediction in list_predictions(body=ListPredictionsRequest(prediction_ids=list(prediction_id)))
            ]

    def refresh(self):
        """Refresh the prediction data from the OrcaCloud"""
        self.__dict__.update(LabelPrediction.get(self.prediction_id).__dict__)

    def inspect(self):
        """Open a UI to inspect the memories used by this prediction"""
        inspect_prediction_result(self)

    def update(self, *, expected_label: int | None = UNSET, tags: set[str] | None = UNSET) -> None:
        """
        Update editable prediction properties.

        Params:
            expected_label: Value to set for the expected label, defaults to `[UNSET]` if not provided.
            tags: Value to replace existing tags with, defaults to `[UNSET]` if not provided.

        Examples:
            Update the expected label:
            >>> prediction.update(expected_label=1)

            Add a new tag:
            >>> prediction.update(tags=prediction.tags | {"new_tag"})

            Remove expected label and tags:
            >>> prediction.update(expected_label=None, tags=None)
        """
        update_prediction(
            prediction_id=self.prediction_id,
            body=UpdatePredictionRequest(
                expected_label=expected_label if expected_label is not UNSET else CLIENT_UNSET,
                tags=[] if tags is None else list(tags) if tags is not UNSET else CLIENT_UNSET,
            ),
        )
        self.refresh()

    def add_tag(self, tag: str) -> None:
        """
        Add a tag to the prediction

        Params:
            tag: Tag to add to the prediction
        """
        self.update(tags=self.tags | {tag})

    def remove_tag(self, tag: str) -> None:
        """
        Remove a tag from the prediction

        Params:
            tag: Tag to remove from the prediction
        """
        self.update(tags=self.tags - {tag})

    def record_feedback(
        self,
        category: str,
        value: bool | float,
        *,
        comment: str | None = None,
    ):
        """
        Record feedback for the prediction.

        We support recording feedback in several categories for each prediction. A
        [`FeedbackCategory`][orca_sdk.telemetry.FeedbackCategory] is created automatically,
        the first time feedback with a new name is recorded. Categories are global across models.
        The value type of the category is inferred from the first recorded value. Subsequent
        feedback for the same category must be of the same type.

        Params:
            category: Name of the category under which to record the feedback.
            value: Feedback value to record, should be `True` for positive feedback and `False` for
                negative feedback or a [`float`][float] between `-1.0` and `+1.0` where negative
                values indicate negative feedback and positive values indicate positive feedback.
            comment: Optional comment to record with the feedback.

        Examples:
            Record whether a suggestion was accepted or rejected:
            >>> prediction.record_feedback("accepted", True)

            Record star rating as normalized continuous score between `-1.0` and `+1.0`:
            >>> prediction.record_feedback("rating", -0.5, comment="2 stars")

        Raises:
            ValueError: If the value does not match previous value types for the category, or is a
                [`float`][float] that is not between `-1.0` and `+1.0`.
        """
        record_prediction_feedback(
            body=[
                _parse_feedback(
                    {"prediction_id": self.prediction_id, "category": category, "value": value, "comment": comment}
                )
            ]
        )
        self.refresh()

    def delete_feedback(self, category: str) -> None:
        """
        Delete prediction feedback for a specific category.

        Params:
            category: Name of the category of the feedback to delete.

        Raises:
            ValueError: If the category is not found.
        """
        record_prediction_feedback(
            body=[PredictionFeedbackRequest(prediction_id=self.prediction_id, category_name=category, value=None)]
        )
        self.refresh()



"""
This file is generated by the openapi-python-client tool via the generate_api_client.py script

It is a customized template from the openapi-python-client tool's default template:
https://github.com/openapi-generators/openapi-python-client/blob/861ef5622f10fc96d240dc9becb0edf94e61446c/openapi_python_client/templates/model.py.jinja

The main change is:
- Fix typing issues
"""

# flake8: noqa: C901

from typing import TYPE_CHECKING, Any, Dict, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.memoryset_cluster_analysis_config import MemorysetClusterAnalysisConfig
    from ..models.memoryset_duplicate_analysis_config import MemorysetDuplicateAnalysisConfig
    from ..models.memoryset_label_analysis_config import MemorysetLabelAnalysisConfig
    from ..models.memoryset_neighbor_analysis_config import MemorysetNeighborAnalysisConfig
    from ..models.memoryset_projection_analysis_config import MemorysetProjectionAnalysisConfig


T = TypeVar("T", bound="MemorysetAnalysisConfigs")


@_attrs_define
class MemorysetAnalysisConfigs:
    """
    Attributes:
        neighbor (Union['MemorysetNeighborAnalysisConfig', None, Unset]):
        label (Union['MemorysetLabelAnalysisConfig', None, Unset]):
        duplicate (Union['MemorysetDuplicateAnalysisConfig', None, Unset]):
        projection (Union['MemorysetProjectionAnalysisConfig', None, Unset]):
        cluster (Union['MemorysetClusterAnalysisConfig', None, Unset]):
    """

    neighbor: Union["MemorysetNeighborAnalysisConfig", None, Unset] = UNSET
    label: Union["MemorysetLabelAnalysisConfig", None, Unset] = UNSET
    duplicate: Union["MemorysetDuplicateAnalysisConfig", None, Unset] = UNSET
    projection: Union["MemorysetProjectionAnalysisConfig", None, Unset] = UNSET
    cluster: Union["MemorysetClusterAnalysisConfig", None, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.memoryset_cluster_analysis_config import MemorysetClusterAnalysisConfig
        from ..models.memoryset_duplicate_analysis_config import MemorysetDuplicateAnalysisConfig
        from ..models.memoryset_label_analysis_config import MemorysetLabelAnalysisConfig
        from ..models.memoryset_neighbor_analysis_config import MemorysetNeighborAnalysisConfig
        from ..models.memoryset_projection_analysis_config import MemorysetProjectionAnalysisConfig

        neighbor: Union[Dict[str, Any], None, Unset]
        if isinstance(self.neighbor, Unset):
            neighbor = UNSET
        elif isinstance(self.neighbor, MemorysetNeighborAnalysisConfig):
            neighbor = self.neighbor.to_dict()
        else:
            neighbor = self.neighbor

        label: Union[Dict[str, Any], None, Unset]
        if isinstance(self.label, Unset):
            label = UNSET
        elif isinstance(self.label, MemorysetLabelAnalysisConfig):
            label = self.label.to_dict()
        else:
            label = self.label

        duplicate: Union[Dict[str, Any], None, Unset]
        if isinstance(self.duplicate, Unset):
            duplicate = UNSET
        elif isinstance(self.duplicate, MemorysetDuplicateAnalysisConfig):
            duplicate = self.duplicate.to_dict()
        else:
            duplicate = self.duplicate

        projection: Union[Dict[str, Any], None, Unset]
        if isinstance(self.projection, Unset):
            projection = UNSET
        elif isinstance(self.projection, MemorysetProjectionAnalysisConfig):
            projection = self.projection.to_dict()
        else:
            projection = self.projection

        cluster: Union[Dict[str, Any], None, Unset]
        if isinstance(self.cluster, Unset):
            cluster = UNSET
        elif isinstance(self.cluster, MemorysetClusterAnalysisConfig):
            cluster = self.cluster.to_dict()
        else:
            cluster = self.cluster

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if neighbor is not UNSET:
            field_dict["neighbor"] = neighbor
        if label is not UNSET:
            field_dict["label"] = label
        if duplicate is not UNSET:
            field_dict["duplicate"] = duplicate
        if projection is not UNSET:
            field_dict["projection"] = projection
        if cluster is not UNSET:
            field_dict["cluster"] = cluster

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: dict[str, Any]) -> T:
        from ..models.memoryset_cluster_analysis_config import MemorysetClusterAnalysisConfig
        from ..models.memoryset_duplicate_analysis_config import MemorysetDuplicateAnalysisConfig
        from ..models.memoryset_label_analysis_config import MemorysetLabelAnalysisConfig
        from ..models.memoryset_neighbor_analysis_config import MemorysetNeighborAnalysisConfig
        from ..models.memoryset_projection_analysis_config import MemorysetProjectionAnalysisConfig

        d = src_dict.copy()

        def _parse_neighbor(data: object) -> Union["MemorysetNeighborAnalysisConfig", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                neighbor_type_0 = MemorysetNeighborAnalysisConfig.from_dict(data)

                return neighbor_type_0
            except:  # noqa: E722
                pass
            return cast(Union["MemorysetNeighborAnalysisConfig", None, Unset], data)

        neighbor = _parse_neighbor(d.pop("neighbor", UNSET))

        def _parse_label(data: object) -> Union["MemorysetLabelAnalysisConfig", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                label_type_0 = MemorysetLabelAnalysisConfig.from_dict(data)

                return label_type_0
            except:  # noqa: E722
                pass
            return cast(Union["MemorysetLabelAnalysisConfig", None, Unset], data)

        label = _parse_label(d.pop("label", UNSET))

        def _parse_duplicate(data: object) -> Union["MemorysetDuplicateAnalysisConfig", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                duplicate_type_0 = MemorysetDuplicateAnalysisConfig.from_dict(data)

                return duplicate_type_0
            except:  # noqa: E722
                pass
            return cast(Union["MemorysetDuplicateAnalysisConfig", None, Unset], data)

        duplicate = _parse_duplicate(d.pop("duplicate", UNSET))

        def _parse_projection(data: object) -> Union["MemorysetProjectionAnalysisConfig", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                projection_type_0 = MemorysetProjectionAnalysisConfig.from_dict(data)

                return projection_type_0
            except:  # noqa: E722
                pass
            return cast(Union["MemorysetProjectionAnalysisConfig", None, Unset], data)

        projection = _parse_projection(d.pop("projection", UNSET))

        def _parse_cluster(data: object) -> Union["MemorysetClusterAnalysisConfig", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                cluster_type_0 = MemorysetClusterAnalysisConfig.from_dict(data)

                return cluster_type_0
            except:  # noqa: E722
                pass
            return cast(Union["MemorysetClusterAnalysisConfig", None, Unset], data)

        cluster = _parse_cluster(d.pop("cluster", UNSET))

        memoryset_analysis_configs = cls(
            neighbor=neighbor,
            label=label,
            duplicate=duplicate,
            projection=projection,
            cluster=cluster,
        )

        memoryset_analysis_configs.additional_properties = d
        return memoryset_analysis_configs

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class FilterItemFieldType2ItemType1(str, Enum):
    CLUSTER = "cluster"
    CURRENT_LABEL_NEIGHBOR_CONFIDENCE = "current_label_neighbor_confidence"
    DUPLICATE_MEMORY_IDS = "duplicate_memory_ids"
    EMBEDDING_2D = "embedding_2d"
    HAS_POTENTIAL_DUPLICATES = "has_potential_duplicates"
    IS_DUPLICATE = "is_duplicate"
    NEIGHBOR_LABEL_LOGITS = "neighbor_label_logits"
    NEIGHBOR_PREDICTED_LABEL = "neighbor_predicted_label"
    NEIGHBOR_PREDICTED_LABEL_AMBIGUITY = "neighbor_predicted_label_ambiguity"
    NEIGHBOR_PREDICTED_LABEL_CONFIDENCE = "neighbor_predicted_label_confidence"
    NEIGHBOR_PREDICTED_LABEL_MATCHES_CURRENT_LABEL = "neighbor_predicted_label_matches_current_label"
    NORMALIZED_NEIGHBOR_LABEL_ENTROPY = "normalized_neighbor_label_entropy"
    POTENTIAL_DUPLICATE_MEMORY_IDS = "potential_duplicate_memory_ids"

    def __str__(self) -> str:
        return str(self.value)



from uuid import uuid4

import pytest

from ._generated_api_client.models import EmbeddingEvaluationResponse, TaskStatus
from .datasource import Datasource


def test_create_datasource(datasource, hf_dataset):
    assert datasource is not None
    assert datasource.name == "test_datasource"
    assert datasource.length == len(hf_dataset)


def test_create_datasource_unauthenticated(unauthenticated, hf_dataset):
    with pytest.raises(ValueError, match="Invalid API key"):
        Datasource.from_hf_dataset("test_datasource", hf_dataset)


def test_create_datasource_already_exists_error(hf_dataset, datasource):
    with pytest.raises(ValueError):
        Datasource.from_hf_dataset("test_datasource", hf_dataset, if_exists="error")


def test_create_datasource_already_exists_return(hf_dataset, datasource):
    returned_dataset = Datasource.from_hf_dataset("test_datasource", hf_dataset, if_exists="open")
    assert returned_dataset is not None
    assert returned_dataset.name == "test_datasource"
    assert returned_dataset.length == len(hf_dataset)


def test_open_datasource(datasource):
    fetched_datasource = Datasource.open(datasource.name)
    assert fetched_datasource is not None
    assert fetched_datasource.name == datasource.name
    assert fetched_datasource.length == len(datasource)


def test_open_datasource_unauthenticated(datasource, unauthenticated):
    with pytest.raises(ValueError, match="Invalid API key"):
        Datasource.open("test_datasource")


def test_open_datasource_invalid_input():
    with pytest.raises(ValueError, match=r"Invalid input:.*"):
        Datasource.open("not valid id")


def test_open_datasource_not_found():
    with pytest.raises(LookupError):
        Datasource.open(str(uuid4()))


def test_open_datasource_unauthorized(datasource, unauthorized):
    with pytest.raises(LookupError):
        Datasource.open(datasource.id)


def test_all_datasources(datasource):
    datasources = Datasource.all()
    assert len(datasources) > 0
    assert any(datasource.name == datasource.name for datasource in datasources)


def test_all_datasources_unauthenticated(unauthenticated):
    with pytest.raises(ValueError, match="Invalid API key"):
        Datasource.all()


def test_drop_datasource(hf_dataset):
    Datasource.from_hf_dataset("datasource_to_delete", hf_dataset)
    assert Datasource.exists("datasource_to_delete")
    Datasource.drop("datasource_to_delete")
    assert not Datasource.exists("datasource_to_delete")


def test_drop_datasource_unauthenticated(datasource, unauthenticated):
    with pytest.raises(ValueError, match="Invalid API key"):
        Datasource.drop(datasource.id)


def test_drop_datasource_not_found():
    with pytest.raises(LookupError):
        Datasource.drop(str(uuid4()))
    # ignores error if specified
    Datasource.drop(str(uuid4()), if_not_exists="ignore")


def test_drop_datasource_unauthorized(datasource, unauthorized):
    with pytest.raises(LookupError):
        Datasource.drop(datasource.id)


def test_drop_datasource_invalid_input():
    with pytest.raises(ValueError, match=r"Invalid input:.*"):
        Datasource.drop("not valid id")



from enum import Enum


class ColumnType(str, Enum):
    BOOL = "BOOL"
    ENUM = "ENUM"
    FLOAT = "FLOAT"
    IMAGE = "IMAGE"
    INT = "INT"
    OTHER = "OTHER"
    STRING = "STRING"

    def __str__(self) -> str:
        return str(self.value)



"""
This file is generated by the openapi-python-client tool via the generate_api_client.py script

It is a customized template from the openapi-python-client tool's default template:
https://github.com/openapi-generators/openapi-python-client/blob/861ef5622f10fc96d240dc9becb0edf94e61446c/openapi_python_client/templates/model.py.jinja

The main change is:
- Fix typing issues
"""

# flake8: noqa: C901

from typing import TYPE_CHECKING, Any, List, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.cluster_metrics import ClusterMetrics


T = TypeVar("T", bound="MemorysetClusterMetrics")


@_attrs_define
class MemorysetClusterMetrics:
    """
    Attributes:
        cluster_metrics (List['ClusterMetrics']):
        num_outliers (int):
        num_clusters (int):
    """

    cluster_metrics: List["ClusterMetrics"]
    num_outliers: int
    num_clusters: int
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        cluster_metrics = []
        for cluster_metrics_item_data in self.cluster_metrics:
            cluster_metrics_item = cluster_metrics_item_data.to_dict()
            cluster_metrics.append(cluster_metrics_item)

        num_outliers = self.num_outliers

        num_clusters = self.num_clusters

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "cluster_metrics": cluster_metrics,
                "num_outliers": num_outliers,
                "num_clusters": num_clusters,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: dict[str, Any]) -> T:
        from ..models.cluster_metrics import ClusterMetrics

        d = src_dict.copy()
        cluster_metrics = []
        _cluster_metrics = d.pop("cluster_metrics")
        for cluster_metrics_item_data in _cluster_metrics:
            cluster_metrics_item = ClusterMetrics.from_dict(cluster_metrics_item_data)

            cluster_metrics.append(cluster_metrics_item)

        num_outliers = d.pop("num_outliers")

        num_clusters = d.pop("num_clusters")

        memoryset_cluster_metrics = cls(
            cluster_metrics=cluster_metrics,
            num_outliers=num_outliers,
            num_clusters=num_clusters,
        )

        memoryset_cluster_metrics.additional_properties = d
        return memoryset_cluster_metrics

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""
This file is generated by the openapi-python-client tool via the generate_api_client.py script

It is a customized template from the openapi-python-client tool's default template:
https://github.com/openapi-generators/openapi-python-client/blob/861ef5622f10fc96d240dc9becb0edf94e61446c/openapi_python_client/templates/errors.py.jinja

The main change is:
- Returning the appropriate python exception for Orca API's error responses
"""

import httpx


def get_error_for_response(response: httpx.Response):
    """Raise an appropriate exception based on the response status code and content"""
    if response.status_code == 401:
        return ValueError("Invalid API key")
    elif response.status_code == 403:
        return PermissionError(response.json()["reason"])
    elif response.status_code == 404:
        resource = response.json()["resource"]
        if resource is not None:
            return LookupError(f"The {resource} you are looking for does not exist")
        else:
            return RuntimeError(f"Unknown API route: {response.url}")
    elif response.status_code == 405:
        return RuntimeError(f"Unknown method {response.request.method} for API route: {response.url}")
    elif response.status_code == 409:
        return RuntimeError(response.json()["constraint"])
    elif response.status_code == 422:
        issues = [f"{issue['loc'][-1]}: {issue['msg']}" for issue in response.json()["validation_issues"]]
        return ValueError("Invalid input:\n\t" + "\n\t".join(issues))
    elif response.status_code == 500:
        return RuntimeError(f"Unexpected server error: {response.json()['message']}")
    elif response.status_code == 503:
        return RuntimeError("Orca API is currently unavailable, please try again later")
    else:
        return RuntimeError(f"Unexpected status code: {response.status_code}")



from uuid import uuid4

from ..credentials import OrcaCredentials
from .auth import _create_api_key, _delete_api_key, _delete_org, _list_api_keys


def test_list_api_keys(org_id):
    assert len(_list_api_keys(org_id)) >= 1


def test_create_api_key(org_id):
    name = f"test-{uuid4().hex[:8]}"
    api_key = _create_api_key(org_id=org_id, name=name)
    assert api_key is not None
    assert name in [api_key.name for api_key in OrcaCredentials.list_api_keys()]


def test_delete_api_key(org_id):
    name = f"test-{uuid4().hex[:8]}"
    api_key = _create_api_key(org_id=org_id, name=name)
    assert api_key is not None
    assert name in [api_key.name for api_key in OrcaCredentials.list_api_keys()]
    _delete_api_key(org_id=org_id, name=name)
    assert name not in [api_key.name for api_key in OrcaCredentials.list_api_keys()]


def test_delete_org(other_org_id):
    _create_api_key(org_id=other_org_id, name="test")
    assert len(_list_api_keys(other_org_id)) >= 1
    _delete_org(other_org_id)
    assert len(_list_api_keys(other_org_id)) == 0



import time

from tqdm.auto import tqdm

from .._generated_api_client.api import abort_task as _abort_task
from .._generated_api_client.api import get_task_status_task
from .._generated_api_client.api import list_tasks as _list_tasks
from .._generated_api_client.models import Task, TaskStatus, TaskStatusInfo

task_config = {
    "retry_interval": 3,
    "show_progress": True,
    "max_wait": 60 * 60,
}


def set_task_config(
    retry_interval: int | None = None,
    show_progress: bool | None = None,
    max_wait: int | None = None,
) -> None:
    if retry_interval is not None:
        task_config["retry_interval"] = retry_interval
    if show_progress is not None:
        task_config["show_progress"] = show_progress
    if max_wait is not None:
        task_config["max_wait"] = max_wait


def wait_for_task(task_id: str, description: str | None = None, show_progress: bool = True) -> None:
    start_time = time.time()
    pbar = None
    steps_total = None
    show_progress = show_progress and task_config["show_progress"]
    while True:
        task_status = get_task_status_task(task_id)

        # setup progress bar if steps total is known
        if task_status.steps_total is not None and steps_total is None:
            steps_total = task_status.steps_total
        if not pbar and steps_total is not None and show_progress:
            pbar = tqdm(total=steps_total, desc=description)

        # return if task is complete
        if task_status.status == TaskStatus.COMPLETED:
            if pbar:
                pbar.update(steps_total - pbar.n)
                pbar.close()
            return

        # raise error if task failed
        if task_status.status == TaskStatus.FAILED:
            raise RuntimeError(f"Task failed with: {task_status.exception}")

        # raise error if task timed out
        if (time.time() - start_time) > task_config["max_wait"]:
            raise RuntimeError(f"Task {task_id} timed out after {task_config['max_wait']}s")

        # update progress bar
        if pbar and task_status.steps_completed is not None:
            pbar.update(task_status.steps_completed - pbar.n)

        # sleep before retrying
        time.sleep(task_config["retry_interval"])


def abort_task(task_id: str) -> TaskStatusInfo:
    _abort_task(task_id)
    return get_task_status_task(task_id)


def list_tasks() -> list[Task]:
    return _list_tasks()



"""
This file is generated by the openapi-python-client tool via the generate_api_client.py script

It is a customized template from the openapi-python-client tool's default template:
https://github.com/openapi-generators/openapi-python-client/blob/861ef5622f10fc96d240dc9becb0edf94e61446c/openapi_python_client/templates/model.py.jinja

The main change is:
- Fix typing issues
"""

# flake8: noqa: C901

from enum import Enum
from typing import Any, List, Literal, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.telemetry_field_type_0_item_type_2 import TelemetryFieldType0ItemType2
from ..models.telemetry_sort_options_direction import TelemetrySortOptionsDirection

T = TypeVar("T", bound="TelemetrySortOptions")


@_attrs_define
class TelemetrySortOptions:
    """
    Attributes:
        field (Union[List[Union[Literal['count'], Literal['lookup']]], List[Union[Literal['feedback_metrics'],
            TelemetryFieldType0ItemType2, str]]]):
        direction (TelemetrySortOptionsDirection):
    """

    field: Union[
        List[Union[Literal["count"], Literal["lookup"]]],
        List[Union[Literal["feedback_metrics"], TelemetryFieldType0ItemType2, str]],
    ]
    direction: TelemetrySortOptionsDirection
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        field: Union[List[Union[Literal["count"], Literal["lookup"]]], List[Union[Literal["feedback_metrics"], str]]]
        if isinstance(self.field, list):
            field = []
            for componentsschemas_telemetry_field_type_0_item_data in self.field:
                componentsschemas_telemetry_field_type_0_item: Union[Literal["feedback_metrics"], str]
                if isinstance(componentsschemas_telemetry_field_type_0_item_data, TelemetryFieldType0ItemType2):
                    componentsschemas_telemetry_field_type_0_item = (
                        componentsschemas_telemetry_field_type_0_item_data.value
                        if isinstance(componentsschemas_telemetry_field_type_0_item_data, Enum)
                        else componentsschemas_telemetry_field_type_0_item_data
                    )
                else:
                    componentsschemas_telemetry_field_type_0_item = componentsschemas_telemetry_field_type_0_item_data
                field.append(componentsschemas_telemetry_field_type_0_item)

        direction = self.direction.value if isinstance(self.direction, Enum) else self.direction

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "field": field,
                "direction": direction,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()

        def _parse_field(
            data: object,
        ) -> Union[
            List[Union[Literal["count"], Literal["lookup"]]],
            List[Union[Literal["feedback_metrics"], TelemetryFieldType0ItemType2, str]],
        ]:
            try:
                if not isinstance(data, list):
                    raise TypeError()
                componentsschemas_telemetry_field_type_0 = []
                _componentsschemas_telemetry_field_type_0 = data
                for componentsschemas_telemetry_field_type_0_item_data in _componentsschemas_telemetry_field_type_0:

                    def _parse_componentsschemas_telemetry_field_type_0_item(
                        data: object,
                    ) -> Union[Literal["feedback_metrics"], TelemetryFieldType0ItemType2, str]:
                        componentsschemas_telemetry_field_type_0_item_type_0 = cast(Literal["feedback_metrics"], data)
                        if componentsschemas_telemetry_field_type_0_item_type_0 != "feedback_metrics":
                            raise ValueError(
                                f"/components/schemas/TelemetryField_type_0_item_type_0 must match const 'feedback_metrics', got '{componentsschemas_telemetry_field_type_0_item_type_0}'"
                            )
                        return componentsschemas_telemetry_field_type_0_item_type_0
                        try:
                            if not isinstance(data, str):
                                raise TypeError()
                            componentsschemas_telemetry_field_type_0_item_type_2 = TelemetryFieldType0ItemType2(data)

                            return componentsschemas_telemetry_field_type_0_item_type_2
                        except:  # noqa: E722
                            pass
                        return cast(Union[Literal["feedback_metrics"], TelemetryFieldType0ItemType2, str], data)

                    componentsschemas_telemetry_field_type_0_item = (
                        _parse_componentsschemas_telemetry_field_type_0_item(
                            componentsschemas_telemetry_field_type_0_item_data
                        )
                    )

                    componentsschemas_telemetry_field_type_0.append(componentsschemas_telemetry_field_type_0_item)

                return componentsschemas_telemetry_field_type_0
            except:  # noqa: E722
                pass
            if not isinstance(data, list):
                raise TypeError()
            componentsschemas_telemetry_field_type_1 = []
            _componentsschemas_telemetry_field_type_1 = data
            for componentsschemas_telemetry_field_type_1_item_data in _componentsschemas_telemetry_field_type_1:

                def _parse_componentsschemas_telemetry_field_type_1_item(
                    data: object,
                ) -> Union[Literal["count"], Literal["lookup"]]:
                    componentsschemas_telemetry_field_type_1_item_type_0 = cast(Literal["lookup"], data)
                    if componentsschemas_telemetry_field_type_1_item_type_0 != "lookup":
                        raise ValueError(
                            f"/components/schemas/TelemetryField_type_1_item_type_0 must match const 'lookup', got '{componentsschemas_telemetry_field_type_1_item_type_0}'"
                        )
                    return componentsschemas_telemetry_field_type_1_item_type_0
                    componentsschemas_telemetry_field_type_1_item_type_1 = cast(Literal["count"], data)
                    if componentsschemas_telemetry_field_type_1_item_type_1 != "count":
                        raise ValueError(
                            f"/components/schemas/TelemetryField_type_1_item_type_1 must match const 'count', got '{componentsschemas_telemetry_field_type_1_item_type_1}'"
                        )
                    return componentsschemas_telemetry_field_type_1_item_type_1

                componentsschemas_telemetry_field_type_1_item = _parse_componentsschemas_telemetry_field_type_1_item(
                    componentsschemas_telemetry_field_type_1_item_data
                )

                componentsschemas_telemetry_field_type_1.append(componentsschemas_telemetry_field_type_1_item)

            return componentsschemas_telemetry_field_type_1

        field = _parse_field(d.pop("field"))

        direction = TelemetrySortOptionsDirection(d.pop("direction"))

        telemetry_sort_options = cls(
            field=field,
            direction=direction,
        )

        telemetry_sort_options.additional_properties = d
        return telemetry_sort_options

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""
This file is generated by the openapi-python-client tool via the generate_api_client.py script

It is a customized template from the openapi-python-client tool's default template:
https://github.com/openapi-generators/openapi-python-client/blob/861ef5622f10fc96d240dc9becb0edf94e61446c/openapi_python_client/templates/model.py.jinja

The main change is:
- Fix typing issues
"""

# flake8: noqa: C901

from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.filter_item import FilterItem
    from ..models.telemetry_filter_item import TelemetryFilterItem
    from ..models.telemetry_sort_options import TelemetrySortOptions


T = TypeVar("T", bound="TelemetryMemoriesRequest")


@_attrs_define
class TelemetryMemoriesRequest:
    """
    Attributes:
        memoryset_id (str):
        offset (Union[Unset, int]):  Default: 0.
        limit (Union[Unset, int]):  Default: 100.
        filters (Union[Unset, List[Union['FilterItem', 'TelemetryFilterItem']]]):
        sort (Union[List['TelemetrySortOptions'], None, Unset]):
    """

    memoryset_id: str
    offset: Union[Unset, int] = 0
    limit: Union[Unset, int] = 100
    filters: Union[Unset, List[Union["FilterItem", "TelemetryFilterItem"]]] = UNSET
    sort: Union[List["TelemetrySortOptions"], None, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.filter_item import FilterItem

        memoryset_id = self.memoryset_id

        offset = self.offset

        limit = self.limit

        filters: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.filters, Unset):
            filters = []
            for filters_item_data in self.filters:
                filters_item: Dict[str, Any]
                if isinstance(filters_item_data, FilterItem):
                    filters_item = filters_item_data.to_dict()
                else:
                    filters_item = filters_item_data.to_dict()

                filters.append(filters_item)

        sort: Union[List[Dict[str, Any]], None, Unset]
        if isinstance(self.sort, Unset):
            sort = UNSET
        elif isinstance(self.sort, list):
            sort = []
            for sort_type_0_item_data in self.sort:
                sort_type_0_item = sort_type_0_item_data.to_dict()
                sort.append(sort_type_0_item)

        else:
            sort = self.sort

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "memoryset_id": memoryset_id,
            }
        )
        if offset is not UNSET:
            field_dict["offset"] = offset
        if limit is not UNSET:
            field_dict["limit"] = limit
        if filters is not UNSET:
            field_dict["filters"] = filters
        if sort is not UNSET:
            field_dict["sort"] = sort

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: dict[str, Any]) -> T:
        from ..models.filter_item import FilterItem
        from ..models.telemetry_filter_item import TelemetryFilterItem
        from ..models.telemetry_sort_options import TelemetrySortOptions

        d = src_dict.copy()
        memoryset_id = d.pop("memoryset_id")

        offset = d.pop("offset", UNSET)

        limit = d.pop("limit", UNSET)

        filters = []
        _filters = d.pop("filters", UNSET)
        for filters_item_data in _filters or []:

            def _parse_filters_item(data: object) -> Union["FilterItem", "TelemetryFilterItem"]:
                try:
                    if not isinstance(data, dict):
                        raise TypeError()
                    filters_item_type_0 = FilterItem.from_dict(data)

                    return filters_item_type_0
                except:  # noqa: E722
                    pass
                if not isinstance(data, dict):
                    raise TypeError()
                filters_item_type_1 = TelemetryFilterItem.from_dict(data)

                return filters_item_type_1

            filters_item = _parse_filters_item(filters_item_data)

            filters.append(filters_item)

        def _parse_sort(data: object) -> Union[List["TelemetrySortOptions"], None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                sort_type_0 = []
                _sort_type_0 = data
                for sort_type_0_item_data in _sort_type_0:
                    sort_type_0_item = TelemetrySortOptions.from_dict(sort_type_0_item_data)

                    sort_type_0.append(sort_type_0_item)

                return sort_type_0
            except:  # noqa: E722
                pass
            return cast(Union[List["TelemetrySortOptions"], None, Unset], data)

        sort = _parse_sort(d.pop("sort", UNSET))

        telemetry_memories_request = cls(
            memoryset_id=memoryset_id,
            offset=offset,
            limit=limit,
            filters=filters,
            sort=sort,
        )

        telemetry_memories_request.additional_properties = d
        return telemetry_memories_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""
This file is generated by the openapi-python-client tool via the generate_api_client.py script

It is a customized template from the openapi-python-client tool's default template:
https://github.com/openapi-generators/openapi-python-client/blob/861ef5622f10fc96d240dc9becb0edf94e61446c/openapi_python_client/templates/model.py.jinja

The main change is:
- Fix typing issues
"""

# flake8: noqa: C901

from typing import Any, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="MemoryMetrics")


@_attrs_define
class MemoryMetrics:
    """
    Attributes:
        is_duplicate (Union[Unset, bool]):
        duplicate_memory_ids (Union[Unset, List[str]]):
        has_potential_duplicates (Union[Unset, bool]):
        potential_duplicate_memory_ids (Union[List[str], None, Unset]):
        cluster (Union[Unset, int]):
        embedding_2d (Union[Unset, List[float]]):
        anomaly_score (Union[Unset, float]):
    """

    is_duplicate: Union[Unset, bool] = UNSET
    duplicate_memory_ids: Union[Unset, List[str]] = UNSET
    has_potential_duplicates: Union[Unset, bool] = UNSET
    potential_duplicate_memory_ids: Union[List[str], None, Unset] = UNSET
    cluster: Union[Unset, int] = UNSET
    embedding_2d: Union[Unset, List[float]] = UNSET
    anomaly_score: Union[Unset, float] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        is_duplicate = self.is_duplicate

        duplicate_memory_ids: Union[Unset, List[str]] = UNSET
        if not isinstance(self.duplicate_memory_ids, Unset):
            duplicate_memory_ids = self.duplicate_memory_ids

        has_potential_duplicates = self.has_potential_duplicates

        potential_duplicate_memory_ids: Union[List[str], None, Unset]
        if isinstance(self.potential_duplicate_memory_ids, Unset):
            potential_duplicate_memory_ids = UNSET
        elif isinstance(self.potential_duplicate_memory_ids, list):
            potential_duplicate_memory_ids = self.potential_duplicate_memory_ids

        else:
            potential_duplicate_memory_ids = self.potential_duplicate_memory_ids

        cluster = self.cluster

        embedding_2d: Union[Unset, List[float]] = UNSET
        if not isinstance(self.embedding_2d, Unset):
            embedding_2d = []
            for embedding_2d_item_data in self.embedding_2d:
                embedding_2d_item: float
                embedding_2d_item = embedding_2d_item_data
                embedding_2d.append(embedding_2d_item)

        anomaly_score = self.anomaly_score

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if is_duplicate is not UNSET:
            field_dict["is_duplicate"] = is_duplicate
        if duplicate_memory_ids is not UNSET:
            field_dict["duplicate_memory_ids"] = duplicate_memory_ids
        if has_potential_duplicates is not UNSET:
            field_dict["has_potential_duplicates"] = has_potential_duplicates
        if potential_duplicate_memory_ids is not UNSET:
            field_dict["potential_duplicate_memory_ids"] = potential_duplicate_memory_ids
        if cluster is not UNSET:
            field_dict["cluster"] = cluster
        if embedding_2d is not UNSET:
            field_dict["embedding_2d"] = embedding_2d
        if anomaly_score is not UNSET:
            field_dict["anomaly_score"] = anomaly_score

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        is_duplicate = d.pop("is_duplicate", UNSET)

        duplicate_memory_ids = cast(List[str], d.pop("duplicate_memory_ids", UNSET))

        has_potential_duplicates = d.pop("has_potential_duplicates", UNSET)

        def _parse_potential_duplicate_memory_ids(data: object) -> Union[List[str], None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                potential_duplicate_memory_ids_type_0 = cast(List[str], data)

                return potential_duplicate_memory_ids_type_0
            except:  # noqa: E722
                pass
            return cast(Union[List[str], None, Unset], data)

        potential_duplicate_memory_ids = _parse_potential_duplicate_memory_ids(
            d.pop("potential_duplicate_memory_ids", UNSET)
        )

        cluster = d.pop("cluster", UNSET)

        embedding_2d = []
        _embedding_2d = d.pop("embedding_2d", UNSET)
        for embedding_2d_item_data in _embedding_2d or []:

            def _parse_embedding_2d_item(data: object) -> float:
                return cast(float, data)

            embedding_2d_item = _parse_embedding_2d_item(embedding_2d_item_data)

            embedding_2d.append(embedding_2d_item)

        anomaly_score = d.pop("anomaly_score", UNSET)

        memory_metrics = cls(
            is_duplicate=is_duplicate,
            duplicate_memory_ids=duplicate_memory_ids,
            has_potential_duplicates=has_potential_duplicates,
            potential_duplicate_memory_ids=potential_duplicate_memory_ids,
            cluster=cluster,
            embedding_2d=embedding_2d,
            anomaly_score=anomaly_score,
        )

        memory_metrics.additional_properties = d
        return memory_metrics

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""
This file is generated by the openapi-python-client tool via the generate_api_client.py script

It is a customized template from the openapi-python-client tool's default template:
https://github.com/openapi-generators/openapi-python-client/blob/861ef5622f10fc96d240dc9becb0edf94e61446c/openapi_python_client/templates/model.py.jinja

The main change is:
- Fix typing issues
"""

# flake8: noqa: C901

from typing import Any, List, Type, TypeVar, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="PredictionFeedbackResult")


@_attrs_define
class PredictionFeedbackResult:
    """
    Attributes:
        prediction_ids (List[str]):
        deleted_feedback_ids (List[str]):
        updated_feedback_ids (List[str]):
        inserted_feedback_ids (List[str]):
        new_category_ids (List[str]):
    """

    prediction_ids: List[str]
    deleted_feedback_ids: List[str]
    updated_feedback_ids: List[str]
    inserted_feedback_ids: List[str]
    new_category_ids: List[str]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        prediction_ids = self.prediction_ids

        deleted_feedback_ids = self.deleted_feedback_ids

        updated_feedback_ids = self.updated_feedback_ids

        inserted_feedback_ids = self.inserted_feedback_ids

        new_category_ids = self.new_category_ids

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "prediction_ids": prediction_ids,
                "deleted_feedback_ids": deleted_feedback_ids,
                "updated_feedback_ids": updated_feedback_ids,
                "inserted_feedback_ids": inserted_feedback_ids,
                "new_category_ids": new_category_ids,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        prediction_ids = cast(List[str], d.pop("prediction_ids"))

        deleted_feedback_ids = cast(List[str], d.pop("deleted_feedback_ids"))

        updated_feedback_ids = cast(List[str], d.pop("updated_feedback_ids"))

        inserted_feedback_ids = cast(List[str], d.pop("inserted_feedback_ids"))

        new_category_ids = cast(List[str], d.pop("new_category_ids"))

        prediction_feedback_result = cls(
            prediction_ids=prediction_ids,
            deleted_feedback_ids=deleted_feedback_ids,
            updated_feedback_ids=updated_feedback_ids,
            inserted_feedback_ids=inserted_feedback_ids,
            new_category_ids=new_category_ids,
        )

        prediction_feedback_result.additional_properties = d
        return prediction_feedback_result

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""
This file is generated by the openapi-python-client tool via the generate_api_client.py script

It is a customized template from the openapi-python-client tool's default template:
https://github.com/openapi-generators/openapi-python-client/blob/861ef5622f10fc96d240dc9becb0edf94e61446c/openapi_python_client/templates/model.py.jinja

The main change is:
- Fix typing issues
"""

# flake8: noqa: C901

from typing import TYPE_CHECKING, Any, Dict, Type, TypeVar, Union, cast

from attrs import define as _attrs_define

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.labeled_memory_insert_metadata import LabeledMemoryInsertMetadata


T = TypeVar("T", bound="LabeledMemoryInsert")


@_attrs_define
class LabeledMemoryInsert:
    """Model of fields required for inserting a labeled memory

    Attributes:
        value (str):
        label (int):
        memory_id (Union[None, Unset, str]):
        metadata (Union[Unset, LabeledMemoryInsertMetadata]):
        source_id (Union[None, Unset, str]):
    """

    value: str
    label: int
    memory_id: Union[None, Unset, str] = UNSET
    metadata: Union[Unset, "LabeledMemoryInsertMetadata"] = UNSET
    source_id: Union[None, Unset, str] = UNSET

    def to_dict(self) -> dict[str, Any]:
        value: str
        value = self.value

        label = self.label

        memory_id: Union[None, Unset, str]
        if isinstance(self.memory_id, Unset):
            memory_id = UNSET
        else:
            memory_id = self.memory_id

        metadata: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.metadata, Unset):
            metadata = self.metadata.to_dict()

        source_id: Union[None, Unset, str]
        if isinstance(self.source_id, Unset):
            source_id = UNSET
        else:
            source_id = self.source_id

        field_dict: dict[str, Any] = {}
        field_dict.update(
            {
                "value": value,
                "label": label,
            }
        )
        if memory_id is not UNSET:
            field_dict["memory_id"] = memory_id
        if metadata is not UNSET:
            field_dict["metadata"] = metadata
        if source_id is not UNSET:
            field_dict["source_id"] = source_id

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: dict[str, Any]) -> T:
        from ..models.labeled_memory_insert_metadata import LabeledMemoryInsertMetadata

        d = src_dict.copy()

        def _parse_value(data: object) -> str:
            return cast(str, data)

        value = _parse_value(d.pop("value"))

        label = d.pop("label")

        def _parse_memory_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        memory_id = _parse_memory_id(d.pop("memory_id", UNSET))

        _metadata = d.pop("metadata", UNSET)
        metadata: Union[Unset, LabeledMemoryInsertMetadata]
        if isinstance(_metadata, Unset):
            metadata = UNSET
        else:
            metadata = LabeledMemoryInsertMetadata.from_dict(_metadata)

        def _parse_source_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        source_id = _parse_source_id(d.pop("source_id", UNSET))

        labeled_memory_insert = cls(
            value=value,
            label=label,
            memory_id=memory_id,
            metadata=metadata,
            source_id=source_id,
        )

        return labeled_memory_insert



"""
This file is generated by the openapi-python-client tool via the generate_api_client.py script

It is a customized template from the openapi-python-client tool's default template:
https://github.com/openapi-generators/openapi-python-client/blob/861ef5622f10fc96d240dc9becb0edf94e61446c/openapi_python_client/templates/model.py.jinja

The main change is:
- Fix typing issues
"""

# flake8: noqa: C901

import datetime
from enum import Enum
from typing import Any, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.feedback_type import FeedbackType

T = TypeVar("T", bound="PredictionFeedback")


@_attrs_define
class PredictionFeedback:
    """
    Attributes:
        prediction_id (str):
        category_name (str):
        value (Union[bool, float]):
        comment (Union[None, str]):
        id (str):
        org_id (str):
        category_id (str):
        category_type (FeedbackType):
        created_at (datetime.datetime):
        updated_at (datetime.datetime):
    """

    prediction_id: str
    category_name: str
    value: Union[bool, float]
    comment: Union[None, str]
    id: str
    org_id: str
    category_id: str
    category_type: FeedbackType
    created_at: datetime.datetime
    updated_at: datetime.datetime
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        prediction_id = self.prediction_id

        category_name = self.category_name

        value: Union[bool, float]
        value = self.value

        comment: Union[None, str]
        comment = self.comment

        id = self.id

        org_id = self.org_id

        category_id = self.category_id

        category_type = self.category_type.value if isinstance(self.category_type, Enum) else self.category_type

        created_at = self.created_at.isoformat()

        updated_at = self.updated_at.isoformat()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "prediction_id": prediction_id,
                "category_name": category_name,
                "value": value,
                "comment": comment,
                "id": id,
                "org_id": org_id,
                "category_id": category_id,
                "category_type": category_type,
                "created_at": created_at,
                "updated_at": updated_at,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        prediction_id = d.pop("prediction_id")

        category_name = d.pop("category_name")

        def _parse_value(data: object) -> Union[bool, float]:
            return cast(Union[bool, float], data)

        value = _parse_value(d.pop("value"))

        def _parse_comment(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        comment = _parse_comment(d.pop("comment"))

        id = d.pop("id")

        org_id = d.pop("org_id")

        category_id = d.pop("category_id")

        category_type = FeedbackType(d.pop("category_type"))

        created_at = isoparse(d.pop("created_at"))

        updated_at = isoparse(d.pop("updated_at"))

        prediction_feedback = cls(
            prediction_id=prediction_id,
            category_name=category_name,
            value=value,
            comment=comment,
            id=id,
            org_id=org_id,
            category_id=category_id,
            category_type=category_type,
            created_at=created_at,
            updated_at=updated_at,
        )

        prediction_feedback.additional_properties = d
        return prediction_feedback

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""
This file is generated by the openapi-python-client tool via the generate_api_client.py script

It is a customized template from the openapi-python-client tool's default template:
https://github.com/openapi-generators/openapi-python-client/blob/861ef5622f10fc96d240dc9becb0edf94e61446c/openapi_python_client/templates/model.py.jinja

The main change is:
- Fix typing issues
"""

# flake8: noqa: C901

import datetime
from enum import Enum
from typing import TYPE_CHECKING, Any, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.pretrained_embedding_model_name import PretrainedEmbeddingModelName
from ..models.task_status import TaskStatus

if TYPE_CHECKING:
    from ..models.memoryset_metrics import MemorysetMetrics


T = TypeVar("T", bound="LabeledMemorysetMetadata")


@_attrs_define
class LabeledMemorysetMetadata:
    """
    Attributes:
        id (str):
        org_id (str):
        collection_name (str):
        name (str):
        description (Union[None, str]):
        length (int):
        pretrained_embedding_model_name (Union[None, PretrainedEmbeddingModelName]):
        finetuned_embedding_model_id (Union[None, str]):
        label_names (List[str]):
        created_at (datetime.datetime):
        updated_at (datetime.datetime):
        insertion_task_id (str):
        insertion_status (TaskStatus): Status of task in the task queue
        metrics (MemorysetMetrics):
    """

    id: str
    org_id: str
    collection_name: str
    name: str
    description: Union[None, str]
    length: int
    pretrained_embedding_model_name: Union[None, PretrainedEmbeddingModelName]
    finetuned_embedding_model_id: Union[None, str]
    label_names: List[str]
    created_at: datetime.datetime
    updated_at: datetime.datetime
    insertion_task_id: str
    insertion_status: TaskStatus
    metrics: "MemorysetMetrics"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        id = self.id

        org_id = self.org_id

        collection_name = self.collection_name

        name = self.name

        description: Union[None, str]
        description = self.description

        length = self.length

        pretrained_embedding_model_name: Union[None, str]
        if isinstance(self.pretrained_embedding_model_name, PretrainedEmbeddingModelName):
            pretrained_embedding_model_name = (
                self.pretrained_embedding_model_name.value
                if isinstance(self.pretrained_embedding_model_name, Enum)
                else self.pretrained_embedding_model_name
            )
        else:
            pretrained_embedding_model_name = self.pretrained_embedding_model_name

        finetuned_embedding_model_id: Union[None, str]
        finetuned_embedding_model_id = self.finetuned_embedding_model_id

        label_names = self.label_names

        created_at = self.created_at.isoformat()

        updated_at = self.updated_at.isoformat()

        insertion_task_id = self.insertion_task_id

        insertion_status = (
            self.insertion_status.value if isinstance(self.insertion_status, Enum) else self.insertion_status
        )

        metrics = self.metrics.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "org_id": org_id,
                "collection_name": collection_name,
                "name": name,
                "description": description,
                "length": length,
                "pretrained_embedding_model_name": pretrained_embedding_model_name,
                "finetuned_embedding_model_id": finetuned_embedding_model_id,
                "label_names": label_names,
                "created_at": created_at,
                "updated_at": updated_at,
                "insertion_task_id": insertion_task_id,
                "insertion_status": insertion_status,
                "metrics": metrics,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: dict[str, Any]) -> T:
        from ..models.memoryset_metrics import MemorysetMetrics

        d = src_dict.copy()
        id = d.pop("id")

        org_id = d.pop("org_id")

        collection_name = d.pop("collection_name")

        name = d.pop("name")

        def _parse_description(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        description = _parse_description(d.pop("description"))

        length = d.pop("length")

        def _parse_pretrained_embedding_model_name(data: object) -> Union[None, PretrainedEmbeddingModelName]:
            if data is None:
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                pretrained_embedding_model_name_type_0 = PretrainedEmbeddingModelName(data)

                return pretrained_embedding_model_name_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, PretrainedEmbeddingModelName], data)

        pretrained_embedding_model_name = _parse_pretrained_embedding_model_name(
            d.pop("pretrained_embedding_model_name")
        )

        def _parse_finetuned_embedding_model_id(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        finetuned_embedding_model_id = _parse_finetuned_embedding_model_id(d.pop("finetuned_embedding_model_id"))

        label_names = cast(List[str], d.pop("label_names"))

        created_at = isoparse(d.pop("created_at"))

        updated_at = isoparse(d.pop("updated_at"))

        insertion_task_id = d.pop("insertion_task_id")

        insertion_status = TaskStatus(d.pop("insertion_status"))

        metrics = MemorysetMetrics.from_dict(d.pop("metrics"))

        labeled_memoryset_metadata = cls(
            id=id,
            org_id=org_id,
            collection_name=collection_name,
            name=name,
            description=description,
            length=length,
            pretrained_embedding_model_name=pretrained_embedding_model_name,
            finetuned_embedding_model_id=finetuned_embedding_model_id,
            label_names=label_names,
            created_at=created_at,
            updated_at=updated_at,
            insertion_task_id=insertion_task_id,
            insertion_status=insertion_status,
            metrics=metrics,
        )

        labeled_memoryset_metadata.additional_properties = d
        return labeled_memoryset_metadata

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
